import struct
from typing import Optional, Type, NamedTuple

from bxcommon import constants
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.messages.bloxroute.blockchain_network_message import RefreshBlockchainNetworkMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.messages.bloxroute.ack_message import AckMessage
from bxcommon.messages.bloxroute.bdn_performance_stats_message import BdnPerformanceStatsMessage
from bxcommon.messages.bloxroute.block_confirmation_message import BlockConfirmationMessage
from bxcommon.messages.bloxroute.block_holding_message import BlockHoldingMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.broadcast_message import BroadcastMessage
from bxcommon.messages.bloxroute.compressed_block_txs_message import CompressedBlockTxsMessage
from bxcommon.messages.bloxroute.disconnect_relay_peer_message import DisconnectRelayPeerMessage
from bxcommon.messages.bloxroute.get_compressed_block_txs_message import GetCompressedBlockTxsMessage
from bxcommon.messages.bloxroute.get_tx_contents_message import GetTxContentsMessage
from bxcommon.messages.bloxroute.get_txs_message import GetTxsMessage
from bxcommon.messages.bloxroute.hello_message import HelloMessage
from bxcommon.messages.bloxroute.key_message import KeyMessage
from bxcommon.messages.bloxroute.notification_message import NotificationMessage
from bxcommon.messages.bloxroute.ping_message import PingMessage
from bxcommon.messages.bloxroute.pong_message import PongMessage
from bxcommon.messages.bloxroute.routing_update_message import RoutingUpdateMessage
from bxcommon.messages.bloxroute.transaction_cleanup_message import TransactionCleanupMessage
from bxcommon.messages.bloxroute.tx_contents_message import TxContentsMessage
from bxcommon.messages.bloxroute.tx_message import TxMessage
from bxcommon.messages.bloxroute.tx_service_sync_blocks_short_ids_message import \
    TxServiceSyncBlocksShortIdsMessage
from bxcommon.messages.bloxroute.tx_service_sync_complete_message import \
    TxServiceSyncCompleteMessage
from bxcommon.messages.bloxroute.tx_service_sync_req_message import TxServiceSyncReqMessage
from bxcommon.messages.bloxroute.tx_service_sync_txs_message import TxServiceSyncTxsMessage
from bxcommon.messages.bloxroute.txs_message import TxsMessage
from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.utils import crypto, uuid_pack
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxcommon.utils.object_hash import ConcatHash, Sha256Hash


class BroadcastMessagePreview(NamedTuple):
    is_full_header: bool
    block_hash: Optional[Sha256Hash]
    broadcast_type: Optional[BroadcastMessageType]
    message_id: Optional[ConcatHash]
    network_num: Optional[int]
    source_id: Optional[str]
    payload_length: Optional[int]


class _BloxrouteMessageFactory(AbstractMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        BloxrouteMessageType.HELLO: HelloMessage,
        BloxrouteMessageType.ACK: AckMessage,
        BloxrouteMessageType.PING: PingMessage,
        BloxrouteMessageType.PONG: PongMessage,
        BloxrouteMessageType.BROADCAST: BroadcastMessage,
        BloxrouteMessageType.TRANSACTION: TxMessage,
        BloxrouteMessageType.GET_TRANSACTIONS: GetTxsMessage,
        BloxrouteMessageType.TRANSACTIONS: TxsMessage,
        BloxrouteMessageType.GET_TX_CONTENTS: GetTxContentsMessage,
        BloxrouteMessageType.TX_CONTENTS: TxContentsMessage,
        BloxrouteMessageType.KEY: KeyMessage,
        BloxrouteMessageType.BLOCK_HOLDING: BlockHoldingMessage,
        BloxrouteMessageType.DISCONNECT_RELAY_PEER: DisconnectRelayPeerMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_REQ: TxServiceSyncReqMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_BLOCKS_SHORT_IDS: TxServiceSyncBlocksShortIdsMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_TXS: TxServiceSyncTxsMessage,
        BloxrouteMessageType.TX_SERVICE_SYNC_COMPLETE: TxServiceSyncCompleteMessage,
        BloxrouteMessageType.BLOCK_CONFIRMATION: BlockConfirmationMessage,
        BloxrouteMessageType.TRANSACTION_CLEANUP: TransactionCleanupMessage,
        BloxrouteMessageType.NOTIFICATION: NotificationMessage,
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: BdnPerformanceStatsMessage,
        BloxrouteMessageType.REFRESH_BLOCKCHAIN_NETWORK: RefreshBlockchainNetworkMessage,
        BloxrouteMessageType.GET_COMPRESSED_BLOCK_TXS: GetCompressedBlockTxsMessage,
        BloxrouteMessageType.COMPRESSED_BLOCK_TXS: CompressedBlockTxsMessage,
        BloxrouteMessageType.ROUTING_UPDATE: RoutingUpdateMessage,
    }

    def __init__(self) -> None:
        super(_BloxrouteMessageFactory, self).__init__()
        self.message_type_mapping = self._MESSAGE_TYPE_MAPPING

    def get_base_message_type(self) -> Type[AbstractMessage]:
        return AbstractBloxrouteMessage

    def get_broadcast_message_preview(self, input_buffer: InputBuffer) -> BroadcastMessagePreview:
        """
        Peeks the hash and network number from hashed messages.
        Currently, only Broadcast messages are supported here.
        :param input_buffer
        :return: is full header, message hash, network number, source id, payload length
        """
        broadcast_header_length = (
            self.base_message_type.HEADER_LENGTH
            + AbstractBroadcastMessage.PAYLOAD_LENGTH
            - constants.CONTROL_FLAGS_LEN
        )
        is_full_header = input_buffer.length >= broadcast_header_length
        if not is_full_header:
            return BroadcastMessagePreview(False, None, None, None, None, None, None)
        else:
            _is_full_message, _command, payload_length = self.get_message_header_preview_from_input_buffer(input_buffer)

            broadcast_header = input_buffer.peek_message(broadcast_header_length)

            offset = self.base_message_type.HEADER_LENGTH

            block_hash = broadcast_header[offset:offset + crypto.SHA256_HASH_LEN]
            block_hash_with_network_num = broadcast_header[offset:
                                                           offset + crypto.SHA256_HASH_LEN + constants.NETWORK_NUM_LEN]
            offset += crypto.SHA256_HASH_LEN

            network_num, = struct.unpack_from("<L", broadcast_header[offset:offset + constants.NETWORK_NUM_LEN])
            offset += constants.NETWORK_NUM_LEN

            source_id = uuid_pack.from_bytes(
                struct.unpack_from("<16s", broadcast_header[offset:offset + constants.NODE_ID_SIZE_IN_BYTES])[0])
            offset += constants.NODE_ID_SIZE_IN_BYTES

            broadcast_type_bytearray = broadcast_header[offset:offset + constants.BROADCAST_TYPE_LEN]
            broadcast_type_in_str = struct.unpack_from(
                "<4s", broadcast_type_bytearray
            )[0].decode(constants.DEFAULT_TEXT_ENCODING)
            broadcast_type = BroadcastMessageType(broadcast_type_in_str)
            message_id = ConcatHash(bytearray(block_hash_with_network_num) + broadcast_type_bytearray, 0)

            return BroadcastMessagePreview(is_full_header, Sha256Hash(block_hash), broadcast_type, message_id,
                                           network_num, source_id, payload_length)

    def __repr__(self):
        return f"{self.__class__.__name__}; message_type_mapping: {self.message_type_mapping}"


bloxroute_message_factory = _BloxrouteMessageFactory()



import os
import ssl
import sys
import urllib.request
from ssl import Purpose
from typing import Optional
from os import path
from pathlib import Path

import websockets

from bloxroute_cli.provider.ws_provider import WsProvider
from bxcommon.models.node_type import NodeType
from bxcommon import constants
from bxutils import constants as utils_constants

CLOUD_WEBSOCKETS_URL = "wss://eth.feed.blxrbdn.com:28333"
CA_CERT_URL = "https://certificates.blxrbdn.com/ca/ca_cert.pem"
DEFAULT_NODE_NAME = NodeType.EXTERNAL_GATEWAY.name.lower()


def _get_default_data_path():
    home_path = str(Path.home())
    if sys.platform == constants.PLATFORM_MAC:
        return path.join(home_path, "Library", "bloXroute")
    else:
        return path.join(home_path, "bloxroute")


class CloudWssProvider(WsProvider):
    def __init__(
        self,
        *,
        ssl_dir: Optional[str] = None,
        ssl_certificate_location: Optional[str] = None,
        ssl_key_location: Optional[str] = None,
        ca_file: Optional[str] = None,
        ca_url: str = CA_CERT_URL,
        ws_uri: str = CLOUD_WEBSOCKETS_URL,
        node_type: str = DEFAULT_NODE_NAME
    ):
        if ssl_dir is None:
            ssl_dir = os.path.join(
                _get_default_data_path(),
                utils_constants.SSL_FOLDER,
                node_type,
                "registration_only"
            )
        if ssl_certificate_location is None:
            ssl_certificate_location = os.path.join(
                ssl_dir, f"{node_type}_cert.pem"
            )
        if ssl_key_location is None:
            ssl_key_location = os.path.join(
                ssl_dir, f"{node_type}_key.pem"
            )
        if ca_file is None:
            with urllib.request.urlopen(ca_url) as cert_file:
                contents = cert_file.read().decode("utf-8")
                context = ssl.create_default_context(
                    Purpose.SERVER_AUTH,
                    cadata=contents,
                )
        else:
            context = ssl.create_default_context(
                Purpose.SERVER_AUTH,
                cafile=ca_file,
            )

        super().__init__(ws_uri)
        context.load_cert_chain(
            ssl_certificate_location, ssl_key_location
        )
        context.check_hostname = False
        self.ssl_context = context

    async def connect_websocket(self) -> websockets.WebSocketClientProtocol:
        return await websockets.connect(self.uri, ssl=self.ssl_context)



import struct
from bxcommon import constants
from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter
from bxgateway.messages.gateway.v1.gateway_hello_message_v1 import GatewayHelloMessageV1
from bxgateway.messages.gateway.gateway_hello_message import GatewayHelloMessage as GatewayHelloMessageV2
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage


class GatewayHelloMessageConverterV1(AbstractMessageConverter):
    def convert_to_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:

        if not isinstance(msg, GatewayHelloMessageV2):
            raise TypeError("GatewayHelloMessage is expected")

        msg_bytes = msg.rawbytes()
        mem_view = memoryview(msg_bytes)

        result_bytes = bytearray(len(msg_bytes) - GatewayHelloMessageV2.ADDITIONAL_LENGTH)
        payload_len = GatewayHelloMessageV1.PAYLOAD_LENGTH

        result_bytes[:] = mem_view[:len(msg_bytes) - GatewayHelloMessageV2.ADDITIONAL_LENGTH]

        return AbstractBloxrouteMessage.initialize_class(GatewayHelloMessageV1, result_bytes, (GatewayMessageType.HELLO, payload_len))

    def convert_from_older_version(self, msg: AbstractInternalMessage) -> AbstractInternalMessage:

        if not isinstance(msg, GatewayHelloMessageV1):
            raise TypeError("GatewayHelloMessageV1 is expected")

        msg_bytes = msg.rawbytes()
        mem_view = memoryview(msg_bytes)

        result_bytes = bytearray(len(msg_bytes) + constants.NODE_ID_SIZE_IN_BYTES)
        payload_len = GatewayHelloMessageV2.PAYLOAD_LENGTH
        off = 0
        result_bytes[off:len(msg_bytes)] = mem_view[:]
        off = len(msg_bytes)
        struct.pack_into("%ss" % constants.NODE_ID_SIZE_IN_BYTES, result_bytes, off, b'')

        return AbstractBloxrouteMessage.initialize_class(GatewayHelloMessageV2, result_bytes, (GatewayMessageType.HELLO, payload_len))


    def convert_first_bytes_to_older_version(self, first_msg_bytes):
        raise NotImplementedError()

    def convert_first_bytes_from_older_version(self, first_msg_bytes):
        raise NotImplementedError()

    def get_message_size_change_to_older_version(self):
        raise NotImplementedError()

    def get_message_size_change_from_older_version(self):
        raise NotImplementedError()


# pyre-fixme[45]: Cannot instantiate abstract class `GatewayHelloMessageConverterV1`.
gateway_hello_message_converter_v1 = GatewayHelloMessageConverterV1()



from typing import TYPE_CHECKING, Callable, Optional, Tuple

from bxcommon.feed.feed import FeedKey
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.rpc_errors import RpcInvalidParams
from bxcommon.feed.feed_manager import FeedManager

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


class UnsubscribeRpcRequest(AbstractRpcRequest["AbstractNode"]):
    help = {
        "params": "[subscription_id]: Subscription ID returned from subscribe call",
        "description": "Unsubscribe from provided subscription ID"
    }

    def __init__(
        self,
        request: BxJsonRpcRequest,
        node: "AbstractNode",
        feed_manager: FeedManager,
        unsubscribe_handler: Callable[[str], Tuple[Optional[FeedKey], Optional[str]]]
    ) -> None:
        self.feed_manager = feed_manager
        self.unsubscribe_handler = unsubscribe_handler
        self.subscriber_id = ""
        super().__init__(request, node)
        assert self.subscriber_id != ""

    def validate_params(self) -> None:
        params = self.params
        if (
            not isinstance(params, list)
            or len(params) != 1
            or not isinstance(params[0], str)
        ):
            raise RpcInvalidParams(
                self.request_id,
                "Unsubscribe RPC request params must be a list of length 1."
            )
        self.subscriber_id = params[0]

    async def process_request(self) -> JsonRpcResponse:
        feed_key, account_id = self.unsubscribe_handler(self.subscriber_id)
        if feed_key is None:
            raise RpcInvalidParams(
                self.request_id,
                f"Subscriber {self.subscriber_id} was not found."
            )
        self.feed_manager.unsubscribe_from_feed(feed_key, self.subscriber_id, account_id)
        return JsonRpcResponse(self.request_id, True)



import struct
from typing import Iterator, Tuple

from bxutils.logging.log_level import LogLevel

from bxcommon.utils import convert
from bxcommon.utils.blockchain_utils.btc import btc_common_utils
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway import btc_constants
from bxgateway.messages.btc import btc_messages_util
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType


def _inv_type_to_witness_inv_type(inv_type):
    if inv_type == InventoryType.MSG_TX:
        return InventoryType.MSG_WITNESS_TX

    if inv_type == InventoryType.MSG_BLOCK:
        return InventoryType.MSG_WITNESS_BLOCK

    return inv_type


class InventoryType(object):
    MSG_WITNESS_FLAG = 1 << 30

    MSG_TX = 1
    MSG_BLOCK = 2
    MSG_WITNESS_BLOCK = MSG_BLOCK | MSG_WITNESS_FLAG
    MSG_WITNESS_TX = MSG_TX | MSG_WITNESS_FLAG
    MSG_CMPCT_BLOCK = 4

    @classmethod
    def is_block(cls, inventory_type):
        return inventory_type != 0 and inventory_type != InventoryType.MSG_TX and inventory_type != InventoryType.MSG_WITNESS_TX


class InventoryBtcMessage(BtcMessage):
    def __init__(self, magic=None, inv_vect=None, command=None, request_witness_data=False, buf=None):
        if buf is None:
            buf = bytearray(
                btc_constants.BTC_HDR_COMMON_OFF + 9 + 36 * len(
                    inv_vect))  # we conservatively allocate the max space for varint
            self.buf = buf

            off = btc_constants.BTC_HDR_COMMON_OFF
            off += btc_messages_util.pack_int_to_btc_varint(len(inv_vect), buf, off)

            for inv_item in inv_vect:

                inv_type = inv_item[0]
                if request_witness_data:
                    inv_type = _inv_type_to_witness_inv_type(inv_item[0])

                struct.pack_into("<L", buf, off, inv_type)
                off += 4
                buf[off:off + 32] = inv_item[1].get_big_endian()
                off += 32

            super(InventoryBtcMessage, self).__init__(magic, command, off - btc_constants.BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

    def count(self):
        off = btc_constants.BTC_HDR_COMMON_OFF
        num_items, size = btc_common_utils.btc_varint_to_int(self.buf, off)
        return num_items

    def __iter__(self) -> Iterator[Tuple[int, BtcObjectHash]]:
        off = btc_constants.BTC_HDR_COMMON_OFF
        num_items, size = btc_common_utils.btc_varint_to_int(self.buf, off)
        off += size
        # pyre-fixme[16]: `InventoryBtcMessage` has no attribute `_num_items`.
        self._num_items = num_items

        # pyre-fixme[16]: `InventoryBtcMessage` has no attribute `_items`.
        self._items = list()
        for _ in range(num_items):
            inv_type, = struct.unpack_from("<L", self.buf, off)
            off += 4
            yield (inv_type, BtcObjectHash(buf=self.buf, offset=off, length=btc_constants.BTC_SHA_HASH_LEN))
            off += 32

    def log_level(self):
        return LogLevel.DEBUG

    def __repr__(self):
        return "InventoryBtcMessage<type: {}, length: {}, items: ({})>".format(self.MESSAGE_TYPE, len(self.rawbytes()),
                                                                               self._inv_vector_to_str())

    def _inv_vector_to_str(self):
        return ", ".join(
            map(lambda inv_item: "{}: {}".format(inv_item[0], convert.bytes_to_hex(inv_item[1].binary)), self))


class InvBtcMessage(InventoryBtcMessage):
    MESSAGE_TYPE = BtcMessageType.INVENTORY

    def __init__(self, magic=None, inv_vects=None, buf=None):
        super(InvBtcMessage, self).__init__(magic, inv_vects, self.MESSAGE_TYPE, False, buf)


class GetDataBtcMessage(InventoryBtcMessage):
    MESSAGE_TYPE = BtcMessageType.GET_DATA

    def __init__(self, magic=None, inv_vects=None, request_witness_data=False, buf=None):
        super(GetDataBtcMessage, self).__init__(magic, inv_vects, self.MESSAGE_TYPE, request_witness_data, buf)


class NotFoundBtcMessage(InventoryBtcMessage):
    MESSAGE_TYPE = BtcMessageType.NOT_FOUND

    def __init__(self, magic=None, inv_vects=None, buf=None):
        super(NotFoundBtcMessage, self).__init__(magic, inv_vects, self.MESSAGE_TYPE, False, buf)



from typing import Union, Optional, List, Tuple

from bxcommon.messages.bloxroute import compact_block_short_ids_serializer
from bxcommon.models.node_type import NodeType
from bxcommon.models.transaction_info import TransactionInfo
from bxcommon.services.transaction_service import TransactionService
from bxcommon.utils.blockchain_utils.eth import rlp_utils, eth_common_utils
from bxcommon.utils.object_hash import Sha256Hash
from bxutils import logging
from bxutils.logging import LogLevel

logger = logging.get_logger(__name__)


def log_compressed_block_debug_info(transaction_service: TransactionService,
                                    block_msg_bytes: Union[memoryview, bytearray]):
    if logger.isEnabledFor(LogLevel.TRACE) and transaction_service.node.opts.block_compression_debug:
        network_num = transaction_service.network_num
        protocol = transaction_service.node.opts.blockchain_networks[network_num].protocol

        # TODO implement in a better way
        if protocol.lower() == "ethereum":
            _log_compressed_block_debug_info_eth(transaction_service, block_msg_bytes)


def log_can_decompress_block(node_type: NodeType,
                             block_hash: Sha256Hash,
                             missing_short_ids: Optional[List[Union[int, TransactionInfo]]]):
    if node_type != NodeType.RELAY_BLOCK:
        if missing_short_ids:
            logger.debug("Can decompress block {}: NO", block_hash)
            logger.debug(
                "Block {} has missing compressed transactions: {}",
                block_hash,
                missing_short_ids
            )
        else:
            logger.debug("Can decompress block {}: YES", block_hash)


def _log_compressed_block_debug_info_eth(transaction_service: TransactionService,
                                         block_msg_bytes: Union[memoryview, bytearray]):
    is_block_relay = transaction_service.node.NODE_TYPE == NodeType.RELAY_BLOCK
    block_hash, short_ids, txs_bytes = _parse_block_eth(block_msg_bytes)

    # parse statistics variables
    short_tx_index = 0
    tx_start_index = 0

    tx_index_in_block = 0
    txs_info = []
    missing_short_ids = []

    while True:
        if tx_start_index >= len(txs_bytes):
            break

        short_id = 0
        has_contents = False
        assignmnet_time = 0

        _, tx_itm_len, tx_itm_start = rlp_utils.consume_length_prefix(txs_bytes, tx_start_index)
        tx_bytes = txs_bytes[tx_itm_start:tx_itm_start + tx_itm_len]

        is_full_tx_start = 0
        is_full_tx, is_full_tx_len, = rlp_utils.decode_int(tx_bytes, is_full_tx_start)

        _, tx_content_len, tx_content_start = rlp_utils.consume_length_prefix(
            tx_bytes, is_full_tx_start + is_full_tx_len)
        tx_content_bytes = tx_bytes[tx_content_start:tx_content_start + tx_content_len]

        if is_full_tx:
            tx_hash = Sha256Hash(eth_common_utils.keccak_hash(tx_content_bytes))
        else:
            short_id = short_ids[short_tx_index]
            tx_hash, tx_bytes, _ = transaction_service.get_transaction(short_id)
            has_contents = tx_bytes is not None
            if tx_hash is not None:
                assignmnet_time = transaction_service.get_short_id_assign_time(short_id)
            short_tx_index += 1

        if is_block_relay:
            txs_info.append((tx_index_in_block, not is_full_tx, short_id, tx_hash))
        else:
            txs_info.append((tx_index_in_block, not is_full_tx, short_id, tx_hash, has_contents, assignmnet_time))

        tx_index_in_block += 1
        tx_start_index = tx_itm_start + tx_itm_len

        if not is_full_tx and not has_contents:
            missing_short_ids.append(short_id)

    if is_block_relay:
        log_message = \
            "Block content (from block relay) {} from (index, is compressed, short id, hash is full) : {}"
    else:
        log_message = \
            "Block content (full) {} (index, compressed, short id, hash, has contents, assignment time) : {}"

    logger.debug(
        log_message,
        block_hash,
        ",".join(str(tx_info) for tx_info in txs_info)
    )

    node_type = transaction_service.node.NODE_TYPE
    assert node_type is not None
    log_can_decompress_block(node_type, block_hash, missing_short_ids)


def _parse_block_eth(block_msg_bytes: Union[bytearray, memoryview]) -> Tuple[Sha256Hash, List[int], memoryview]:
    block_msg_bytes = block_msg_bytes if isinstance(block_msg_bytes, memoryview) else memoryview(block_msg_bytes)

    block_offsets = compact_block_short_ids_serializer.get_bx_block_offsets(block_msg_bytes)
    short_ids, _short_ids_bytes_len = compact_block_short_ids_serializer.deserialize_short_ids_from_buffer(
        block_msg_bytes,
        block_offsets.short_id_offset
    )

    block_bytes = block_msg_bytes[block_offsets.block_begin_offset: block_offsets.short_id_offset]

    _, _block_itm_len, block_itm_start = rlp_utils.consume_length_prefix(block_bytes, 0)
    block_itm_bytes = block_bytes[block_itm_start:]

    _, block_hdr_len, block_hdr_start = rlp_utils.consume_length_prefix(block_itm_bytes, 0)
    full_hdr_bytes = block_itm_bytes[0:block_hdr_start + block_hdr_len]

    block_hash_bytes = eth_common_utils.keccak_hash(full_hdr_bytes)
    block_hash = Sha256Hash(block_hash_bytes)

    _, block_txs_len, block_txs_start = rlp_utils.consume_length_prefix(
        block_itm_bytes, block_hdr_start + block_hdr_len
    )
    txs_bytes = block_itm_bytes[block_txs_start:block_txs_start + block_txs_len]

    return block_hash, short_ids, txs_bytes



from abc import ABCMeta, abstractmethod
from asyncio import Task
from typing import Optional, NamedTuple, Any, Callable, Tuple, Dict

from bxcommon.rpc.json_rpc_request import JsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.rpc_errors import RpcError


class SubscriptionNotification(NamedTuple):
    subscription_id: str
    notification: Any


class AbstractProvider(metaclass=ABCMeta):
    """
    Provider interface for code that wants to hook into subscription feed available
    in the websockets interface of bxgateway. An IPC interface should also be
    available soon.

    To make use of this functionality, bxgateway must be started with `--ws True`.
    """

    async def __aenter__(self) -> 'AbstractProvider':
        await self.initialize()
        return self

    def __await__(self):
        yield from self.initialize().__await__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    @abstractmethod
    async def initialize(self) -> None:
        pass

    @abstractmethod
    async def call(
        self,
        request: JsonRpcRequest,
    ) -> JsonRpcResponse:
        pass

    @abstractmethod
    async def subscribe(self, channel: str, options: Optional[Dict[str, Any]] = None) -> str:
        pass

    @abstractmethod
    async def unsubscribe(self, subscription_id: str) -> Tuple[bool, Optional[RpcError]]:
        pass

    @abstractmethod
    def subscribe_with_callback(
        self,
        callback: Callable[[SubscriptionNotification], None],
        channel: str,
        options: Optional[Dict[str, Any]] = None,
    ) -> Task:
        pass

    @abstractmethod
    async def get_next_subscription_notification(self) -> SubscriptionNotification:
        pass

    @abstractmethod
    async def get_next_subscription_notification_by_id(
        self, subscription_id: str
    ) -> SubscriptionNotification:
        pass

    @abstractmethod
    async def close(self) -> None:
        pass



from abc import abstractmethod
from typing import Generic, TypeVar, Any, Dict, List, TYPE_CHECKING, Type
import base64

from bxcommon import constants
from bxcommon.rpc import rpc_constants
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.rpc_errors import RpcParseError, RpcError, RpcInternalError
from bxcommon.rpc.rpc_request_type import RpcRequestType
from bxutils import logging, log_messages
from bxutils.encoding.json_encoder import Case

logger = logging.get_logger(__name__)

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode

Node = TypeVar("Node", bound="AbstractNode")
Req = TypeVar("Req")
Res = TypeVar("Res")


class AbstractRpcHandler(Generic[Node, Req, Res]):
    node: Node
    request_handlers: Dict[RpcRequestType, Type[AbstractRpcRequest[Node]]]
    case: Case

    def __init__(self, node: Node, case: Case = Case.SNAKE) -> None:
        self.node = node
        self.request_handlers = {}
        self.case = case

    async def handle_request(self, request: Req) -> Res:
        try:
            payload = await self.parse_request(request)
        except Exception:
            raise RpcParseError(None, f"Unable to parse the request: {request}")

        rpc_request = BxJsonRpcRequest.from_json(payload)

        # request can be many types, not all of them have "headers"
        headers = getattr(request, "headers", None)
        account_cache_key = ""
        if headers and rpc_constants.AUTHORIZATION_HEADER_KEY in headers:
            request_auth_key = headers[rpc_constants.AUTHORIZATION_HEADER_KEY]
            account_cache_key = base64.b64decode(request_auth_key).decode(constants.DEFAULT_TEXT_ENCODING)
            rpc_request.account_cache_key = account_cache_key

        request_handler = self.get_request_handler(rpc_request)
        try:
            return self.serialize_response(
                await request_handler.process_request()
            )
        except RpcError as e:
            raise e
        except Exception as e:
            logger.exception(
                log_messages.INTERNAL_ERROR_HANDLING_RPC_REQUEST, e, rpc_request, account_cache_key
            )
            raise RpcInternalError(rpc_request.id, "Please contact bloXroute support.")

    async def help(self) -> List[Any]:
        return [
            {
                "method": method.name.lower(),
                "id": "Optional - [unique request identifier string].",
                "params": self.request_handlers[method].help["params"],
                "description": self.request_handlers[method].help["description"],
            }
            for method in self.request_handlers
        ]

    @abstractmethod
    async def parse_request(self, request: Req) -> Dict[str, Any]:
        pass

    @abstractmethod
    def get_request_handler(self, request: BxJsonRpcRequest) -> AbstractRpcRequest:
        pass

    @abstractmethod
    def serialize_response(self, response: JsonRpcResponse) -> Res:
        pass



import struct

from bxcommon import constants
from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.bdn_performance_stats_message import BdnPerformanceStatsMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v10.bdn_performance_stats_message_v10 import BdnPerformanceStatsMessageV10
from bxcommon.messages.versioning.abstract_message_converter import AbstractMessageConverter
from bxcommon.utils.stats import message_utils


class _BdnPerformanceStatsMessageConverterV10(AbstractMessageConverter):
    _MSG_TYPE_TO_OLD_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: BdnPerformanceStatsMessageV10
    }

    _MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING = {
        BloxrouteMessageType.BDN_PERFORMANCE_STATS: BdnPerformanceStatsMessage
    }

    _BASE_LENGTH = (
        AbstractBloxrouteMessage.HEADER_LENGTH
    )

    _INTERVAL_TIMES_BREAKPOINT = (
        _BASE_LENGTH +
        (2 * constants.DOUBLE_SIZE_IN_BYTES)
    )
    _INTERVAL_TIMES_LENGTH = _INTERVAL_TIMES_BREAKPOINT - _BASE_LENGTH

    _FIRST_STATS_SETS_BREAKPOINT = (
        _INTERVAL_TIMES_BREAKPOINT +
        (2 * constants.UL_SHORT_SIZE_IN_BYTES) +
        (2 * constants.UL_INT_SIZE_IN_BYTES)
    )
    _FIRST_STATS_SETS_LENGTH = _FIRST_STATS_SETS_BREAKPOINT - _INTERVAL_TIMES_BREAKPOINT

    _OLD_MESSAGE_LEN = (
        BdnPerformanceStatsMessageV10.MSG_SIZE
    )

    _NEW_MESSAGE_LEN = (
        BdnPerformanceStatsMessageV10.MSG_SIZE +
        constants.UL_SHORT_SIZE_IN_BYTES +
        (5 * constants.UL_INT_SIZE_IN_BYTES) +
        constants.IP_ADDR_SIZE_IN_BYTES +
        (2 * constants.UL_SHORT_SIZE_IN_BYTES)
    )

    _LENGTH_DIFFERENCE = _NEW_MESSAGE_LEN - _OLD_MESSAGE_LEN

    def convert_from_older_version(
        self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        msg_type = msg.MESSAGE_TYPE

        if msg_type not in self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING:
            raise ValueError(
                f"Tried to convert unexpected old message type from v10: {msg_type}"
            )

        new_msg_class = self._MSG_TYPE_TO_NEW_MSG_CLASS_MAPPING[msg_type]
        new_payload_len = msg.payload_len() + self._LENGTH_DIFFERENCE

        default_new_stats = 0
        new_msg_bytes = bytearray(self._NEW_MESSAGE_LEN)
        new_msg_bytes[:self._INTERVAL_TIMES_BREAKPOINT] = msg.rawbytes()[:self._INTERVAL_TIMES_BREAKPOINT]
        off = self._BASE_LENGTH + self._INTERVAL_TIMES_LENGTH

        # memory
        struct.pack_into("<H", new_msg_bytes, off, default_new_stats)
        off += constants.UL_SHORT_SIZE_IN_BYTES

        # single blockchain peer
        struct.pack_into("<H", new_msg_bytes, off, 1)
        off += constants.UL_SHORT_SIZE_IN_BYTES

        # placeholder ip/port
        message_utils.pack_ip_port(new_msg_bytes, off, "0.0.0.0", 0)
        off += constants.IP_ADDR_SIZE_IN_BYTES + constants.UL_SHORT_SIZE_IN_BYTES

        new_msg_bytes[off:off + self._FIRST_STATS_SETS_LENGTH] = \
            msg.rawbytes()[
            self._INTERVAL_TIMES_BREAKPOINT:self._INTERVAL_TIMES_BREAKPOINT + self._FIRST_STATS_SETS_LENGTH
            ]
        off += self._FIRST_STATS_SETS_LENGTH

        struct.pack_into("<I", new_msg_bytes, off, default_new_stats)
        off += constants.UL_INT_SIZE_IN_BYTES
        struct.pack_into("<I", new_msg_bytes, off, default_new_stats)
        off += constants.UL_INT_SIZE_IN_BYTES
        struct.pack_into("<I", new_msg_bytes, off, default_new_stats)
        off += constants.UL_INT_SIZE_IN_BYTES
        struct.pack_into("<I", new_msg_bytes, off, default_new_stats)
        off += constants.UL_INT_SIZE_IN_BYTES
        struct.pack_into("<I", new_msg_bytes, off, default_new_stats)
        off += constants.UL_INT_SIZE_IN_BYTES

        new_msg_bytes[off:] = msg.rawbytes()[self._FIRST_STATS_SETS_BREAKPOINT:]

        return AbstractBloxrouteMessage.initialize_class(
            new_msg_class,
            new_msg_bytes,
            (msg_type, new_payload_len)
        )

    def convert_to_older_version(
            self, msg: AbstractInternalMessage
    ) -> AbstractInternalMessage:
        # Message is one-way from gateway to relay, so converting from new to old version is not necessary.
        raise NotImplementedError

    def convert_first_bytes_to_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_first_bytes_from_older_version(
        self, first_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_to_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def convert_last_bytes_from_older_version(
        self, last_msg_bytes: memoryview
    ) -> memoryview:
        raise NotImplementedError

    def get_message_size_change_to_older_version(self) -> int:
        return -self._LENGTH_DIFFERENCE

    def get_message_size_change_from_older_version(self) -> int:
        return self._LENGTH_DIFFERENCE


bdn_performance_stats_message_converter_v10 = _BdnPerformanceStatsMessageConverterV10()



from typing import Any, Dict, TYPE_CHECKING, Generic, TypeVar

from aiohttp import web
from aiohttp.web import Request, Response
from aiohttp.web_exceptions import HTTPBadRequest, HTTPOk

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.abstract_rpc_handler import AbstractRpcHandler
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.rpc_constants import ContentType
from bxcommon.rpc.rpc_errors import RpcParseError
from bxutils import logging

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


logger = logging.get_logger(__name__)
Node = TypeVar("Node", bound="AbstractNode")


class HttpRpcHandler(Generic[Node], AbstractRpcHandler[Node, Request, Response]):
    def __init__(self, node: Node):
        super().__init__(node)
        self.content_type = ContentType.PLAIN

    def parse_content_type(self, request: Request):
        try:
            # pyre-fixme[16]: Callable `aiohttp.web_request.BaseRequest.headers` has no attribute `__getitem__`
            content_type = request.headers[rpc_constants.CONTENT_TYPE_HEADER_KEY]
        except KeyError:
            pass
        else:
            try:
                content_type_enum = ContentType.from_string(content_type)
                self.content_type = content_type_enum
            except ValueError:
                raise HTTPBadRequest(
                    text=f"Unrecognized content type: {content_type}. "
                         f"Accepted values: {ContentType.PLAIN.value}, {ContentType.JSON.value}"
                )

    async def parse_request(self, request: Request) -> Dict[str, Any]:
        payload = await request.json()
        if not isinstance(payload, dict):
            raise RpcParseError(None, f"Unable to parse the request: {payload}")
        return payload

    def get_request_handler(self, request: BxJsonRpcRequest) -> AbstractRpcRequest:
        # seems to be pyre bug: https://github.com/facebook/pyre-check/issues/267
        # pyre-fixme[45]: Cannot instantiate abstract class `Abstract` with abstract method `run`.
        return self.request_handlers[request.method](request, self.node)

    def serialize_response(self, response: JsonRpcResponse) -> Response:
        if response.error is None:
            status_code = HTTPOk.status_code
        else:
            status_code = HTTPBadRequest.status_code

        if self.content_type == ContentType.PLAIN:
            return web.json_response(
                response.to_jsons(self.case),
                status=status_code,
                content_type=ContentType.PLAIN.value
            )
        else:
            return web.json_response(text=response.to_jsons(self.case), status=status_code)



import functools
from concurrent.futures import Future, CancelledError
from typing import Callable, Any, Optional

from bxcommon import constants
from bxcommon.utils.alarm_queue import AlarmQueue
from bxcommon.utils.concurrency.thread_pool import ThreadPool
from bxutils import log_messages
from bxutils import logging

logger = logging.get_logger(__name__)


class ThreadedRequestService:
    """
    Single point for threaded requests with associated alarms
    """
    thread_pool: ThreadPool

    def __init__(self, name_prefix: str, alarm_queue: AlarmQueue, timeout: int) -> None:
        """
        :param alarm_queue: taken from the node that is using the http service
        :param timeout: timeout for the alarm.
        """
        self.alarm_queue = alarm_queue
        self.thread_pool = ThreadPool(
            constants.THREAD_POOL_WORKER_COUNT, "{}_threaded_request_service".format(name_prefix)
        )
        self.timeout = timeout

    def start(self) -> None:
        self.thread_pool.start()

    def close(self) -> None:
        self.thread_pool.stop()
        self.thread_pool.close()

    def send_threaded_request(
        self,
        request: Callable[..., Any],
        *args: Any,
        done_callback: Optional[Callable[[Future], Any]] = None
    ) -> Future:
        """
        Submit a function to be executed in a separate thread in a thread pool,
        and set up an alarm to verify the correct result of the function

        :param request: function that we need to execute in a separate thread
        :param args: list of arguments for the function
        :param done_callback: Callback when the future is done
        """
        logger.trace("Starting thread for request.")
        task = self.thread_pool.submit(request, *args)
        if done_callback:
            main_thread_callback = functools.partial(
                self.alarm_queue.register_alarm, 0, done_callback
            )
            task.add_done_callback(main_thread_callback)

        self.alarm_queue.register_alarm(
            self.timeout,
            self._threaded_post_alarm,
            task,
            request,
            *args,
            alarm_name=f"threaded_status_check: {str(done_callback)}"
        )
        return task

    def _threaded_post_alarm(self, task: Future, request: Callable[..., None], *args: Any) -> None:
        """
        Callback for the alarm queue.
        Check result of the future task and print a log in case something went wrong
        :param task: future task returned by the thread pool executor
        """
        extra_info = "{}, {}".format(request, args)
        if not task.done():
            if task.running():
                logger.warning(log_messages.THREADED_REQUEST_HAS_LONG_RUNTIME, self.timeout, extra_info)
            else:
                logger.warning(log_messages.THREADED_REQUEST_IS_STALE, extra_info)
                task.cancel()
        else:
            try:
                result = task.result()
                logger.trace("Task {} completed with result: {}", extra_info, result)
            except CancelledError as e:
                logger.error(log_messages.TASK_CANCELLED, task, extra_info, e, exc_info=True)
            except Exception as e:  # pylint: disable=broad-except
                logger.error(log_messages.TASK_FAILED, task, extra_info, e, exc_info=True)



import dataclasses
import time

from dataclasses import dataclass
from collections import deque
from typing import Type, Dict, Deque, Any, TYPE_CHECKING

from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.statistics_service import StatisticsService, StatsIntervalData
from bxgateway import gateway_constants
from bxutils import logging, utils
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


@dataclass
class GatewayTransactionStatInterval(StatsIntervalData):
    new_transactions_received_from_blockchain: int = 0
    duplicate_transactions_received_from_blockchain: int = 0
    new_full_transactions_received_from_relays: int = 0
    new_compact_transactions_received_from_relays: int = 0
    duplicate_full_transactions_received_from_relays: int = 0
    duplicate_compact_transactions_received_from_relays: int = 0
    redundant_transaction_content_messages: int = 0
    short_id_assignments_processed: int = 0
    transaction_tracker: Dict[Sha256Hash, float] = dataclasses.field(default_factory=dict)
    transaction_intervals: Deque[float] = dataclasses.field(default_factory=deque)

    total_bdn_transactions_processed: int = 0
    total_bdn_transactions_process_time_ms: float = 0
    total_bdn_transactions_process_time_before_ext_ms: float = 0
    total_bdn_transactions_process_time_ext_ms: float = 0
    total_bdn_transactions_process_time_after_ext_ms: float = 0

    total_node_transactions_processed: int = 0
    total_node_transactions_process_time_ms: float = 0
    total_node_transactions_process_time_before_broadcast_ms: float = 0
    total_node_transactions_process_time_broadcast_ms: float = 0
    total_node_transactions_process_time_after_broadcast_ms: float = 0

    tx_validation_failed_gas_price: int = 0
    tx_validation_failed_structure: int = 0
    tx_validation_failed_signature: int = 0

    dropped_transactions_from_relay: int = 0

    transactions_bytes_skipped: int = 0


class _GatewayTransactionStatsService(
    StatisticsService[GatewayTransactionStatInterval, "AbstractGatewayNode"]
):
    TRANSACTION_SHORT_ID_ASSIGNED_DONE = -1

    def __init__(
        self,
        interval: int = gateway_constants.GATEWAY_TRANSACTION_STATS_INTERVAL_S,
        look_back: int = gateway_constants.GATEWAY_TRANSACTION_STATS_LOOKBACK,
    ) -> None:
        super(_GatewayTransactionStatsService, self).__init__(
            "GatewayTransactionStats",
            interval,
            look_back,
            reset=True,
            stat_logger=logging.get_logger(LogRecordType.TransactionStats, __name__),
        )

    def get_interval_data_class(self) -> Type[GatewayTransactionStatInterval]:
        return GatewayTransactionStatInterval

    def log_transaction_from_blockchain(self, transaction_hash: Sha256Hash) -> None:
        self.interval_data.new_transactions_received_from_blockchain += 1
        if transaction_hash not in self.interval_data.transaction_tracker:
            self.interval_data.transaction_tracker[transaction_hash] = time.time()

    def log_duplicate_transaction_from_blockchain(self) -> None:
        self.interval_data.duplicate_transactions_received_from_blockchain += 1

    def log_transaction_from_relay(
        self, transaction_hash: Sha256Hash, has_short_id: bool, is_compact: bool = False
    ) -> None:
        if has_short_id and transaction_hash in self.interval_data.transaction_tracker:
            start_time = self.interval_data.transaction_tracker[transaction_hash]
            if start_time != self.TRANSACTION_SHORT_ID_ASSIGNED_DONE:
                self.interval_data.transaction_intervals.append(time.time() - start_time)
                self.interval_data.transaction_tracker[
                    transaction_hash
                ] = self.TRANSACTION_SHORT_ID_ASSIGNED_DONE

        if is_compact:
            self.interval_data.new_compact_transactions_received_from_relays += 1
        else:
            self.interval_data.new_full_transactions_received_from_relays += 1

    def log_short_id_assignment_processed(self) -> None:
        self.interval_data.short_id_assignments_processed += 1

    def log_duplicate_transaction_from_relay(self, is_compact=False) -> None:
        if is_compact:
            self.interval_data.duplicate_compact_transactions_received_from_relays += 1
        else:
            self.interval_data.duplicate_full_transactions_received_from_relays += 1

    def log_dropped_transaction_from_relay(self) -> None:
        interval_data = self.interval_data
        interval_data.dropped_transactions_from_relay += 1

    def log_redundant_transaction_content(self) -> None:
        self.interval_data.redundant_transaction_content_messages += 1

    def log_processed_bdn_transaction(self,
                                      processing_time_ms: float,
                                      processing_time_before_ext_ms: float,
                                      processing_time_ext_ms: float,
                                      processing_time_after_ext_ms: float
                                      ) -> None:
        interval_data = self.interval_data
        interval_data.total_bdn_transactions_processed += 1
        interval_data.total_bdn_transactions_process_time_ms += processing_time_ms
        interval_data.total_bdn_transactions_process_time_before_ext_ms += processing_time_before_ext_ms
        interval_data.total_bdn_transactions_process_time_ext_ms += processing_time_ext_ms
        interval_data.total_bdn_transactions_process_time_after_ext_ms += processing_time_after_ext_ms

    def log_processed_node_transaction(self,
                                       total_duration_ms: float,
                                       duration_before_broadcast_ms: float,
                                       duration_broadcast_ms: float,
                                       duration_set_content_ms: float,
                                       count: int
                                       ) -> None:
        interval_data = self.interval_data
        interval_data.total_node_transactions_processed += count
        interval_data.total_node_transactions_process_time_ms += total_duration_ms
        interval_data.total_node_transactions_process_time_before_broadcast_ms += duration_before_broadcast_ms
        interval_data.total_node_transactions_process_time_broadcast_ms += duration_broadcast_ms
        interval_data.total_node_transactions_process_time_after_broadcast_ms += duration_set_content_ms

    def log_tx_validation_failed_signature(self,) -> None:
        self.interval_data.tx_validation_failed_signature += 1

    def log_tx_validation_failed_structure(self,) -> None:
        self.interval_data.tx_validation_failed_structure += 1

    def log_tx_validation_failed_gas_price(self,) -> None:
        self.interval_data.tx_validation_failed_gas_price += 1

    def log_skipped_transaction_bytes(self, skipped_bytes: int) -> None:
        self.interval_data.transactions_bytes_skipped += skipped_bytes

    def get_info(self) -> Dict[str, Any]:
        node = self.node
        assert node is not None

        if len(self.interval_data.transaction_intervals) > 0:
            min_short_id_assign_time = min(self.interval_data.transaction_intervals)
            max_short_id_assign_time = max(self.interval_data.transaction_intervals)
            avg_short_id_assign_time = sum(self.interval_data.transaction_intervals) / len(
                self.interval_data.transaction_intervals
            )
        else:
            min_short_id_assign_time = 0
            max_short_id_assign_time = 0
            avg_short_id_assign_time = 0

        interval_data = self.interval_data

        return {
            "node_id": node.opts.node_id,
            "new_transactions_received_from_blockchain": interval_data.new_transactions_received_from_blockchain,
            "duplicate_transactions_received_from_blockchain": interval_data.duplicate_transactions_received_from_blockchain,
            "new_full_transactions_received_from_relays": interval_data.new_full_transactions_received_from_relays,
            "new_compact_transactions_received_from_relays": interval_data.new_compact_transactions_received_from_relays,
            "duplicate_full_transactions_received_from_relays": interval_data.duplicate_full_transactions_received_from_relays,
            "duplicate_compact_transactions_received_from_relays": interval_data.duplicate_compact_transactions_received_from_relays,
            "redundant_transaction_content_messages": interval_data.redundant_transaction_content_messages,
            "dropped_transactions_from_relay": interval_data.dropped_transactions_from_relay,
            "short_ids_assignments_processed": interval_data.short_id_assignments_processed,
            "start_time": interval_data.start_time,
            "end_time": interval_data.end_time,
            "min_short_id_assign_time": min_short_id_assign_time,
            "max_short_id_assign_time": max_short_id_assign_time,
            "avg_short_id_assign_time": avg_short_id_assign_time,

            "total_bdn_transactions_processed": interval_data.total_bdn_transactions_processed,
            "total_bdn_transactions_process_time_ms": interval_data.total_bdn_transactions_process_time_ms,
            "average_bdn_tx_process_time_ms":
                utils.safe_divide(
                    interval_data.total_bdn_transactions_process_time_ms,
                    interval_data.total_bdn_transactions_processed),
            "average_bdn_tx_process_time_ext_ms":
                utils.safe_divide(
                    interval_data.total_bdn_transactions_process_time_ext_ms,
                    interval_data.total_bdn_transactions_processed),
            "average_bdn_tx_process_time_before_ext_ms":
                utils.safe_divide(
                    interval_data.total_bdn_transactions_process_time_before_ext_ms,
                    interval_data.total_bdn_transactions_processed),
            "average_bdn_tx_process_time_after_ext_ms":
                utils.safe_divide(
                    interval_data.total_bdn_transactions_process_time_after_ext_ms,
                    interval_data.total_bdn_transactions_processed),

            "total_node_transactions_processed": interval_data.total_node_transactions_processed,
            "total_node_transactions_process_time_ms": interval_data.total_node_transactions_process_time_ms,
            "average_node_tx_process_time_ms":
                utils.safe_divide(
                    interval_data.total_node_transactions_process_time_ms,
                    interval_data.total_node_transactions_processed),
            "average_node_tx_process_time_before_broadcast_ms":
                utils.safe_divide(
                    interval_data.total_node_transactions_process_time_before_broadcast_ms,
                    interval_data.total_node_transactions_processed),
            "average_node_tx_process_time_broadcast_ms":
                utils.safe_divide(
                    interval_data.total_node_transactions_process_time_broadcast_ms,
                    interval_data.total_node_transactions_processed),
            "average_node_tx_process_time_after_broadcast_ms":
                utils.safe_divide(
                    interval_data.total_node_transactions_process_time_after_broadcast_ms,
                    interval_data.total_node_transactions_processed),
            "rejected_signature": interval_data.tx_validation_failed_signature,
            "rejected_structure": interval_data.tx_validation_failed_structure,
            "rejected_gas_price": interval_data.tx_validation_failed_gas_price,
            "transaction_bytes_skipped": interval_data.transactions_bytes_skipped,
            **node._tx_service.get_aggregate_stats(),
        }


gateway_transaction_stats_service = _GatewayTransactionStatsService()



import time
import traceback
import dataclasses

from dataclasses import dataclass
from abc import ABCMeta, abstractmethod
from collections import deque
from datetime import datetime
from threading import Thread, Lock
from typing import Optional, TypeVar, Generic, Deque, Type, Callable, Dict, Any, TYPE_CHECKING

from bxcommon import constants
from bxutils import log_messages
from bxutils import logging
from bxutils.logging import CustomLogger
from bxutils.logging.log_level import LogLevel
from bxutils.logging.log_record_type import LogRecordType

logger = logging.get_logger(__name__)
task_duration_logger = logging.get_logger(LogRecordType.TaskDuration, __name__)

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode


@dataclass
class StatsIntervalData:
    start_time: datetime = dataclasses.field(default_factory=datetime.utcnow)
    end_time: Optional[datetime] = None
    _closed: bool = False

    def close(self):
        self.end_time = datetime.utcnow()
        self._closed = True


T = TypeVar("T", bound=StatsIntervalData)
N = TypeVar("N", bound="AbstractNode")


class StatisticsService(Generic[T, N], metaclass=ABCMeta):
    """
    Abstract class of statistics services.
    """
    history: Deque[T]
    node: Optional[N]
    name: str
    log_level: LogLevel
    logger: CustomLogger
    interval_data: T
    interval: int
    reset: bool

    def __init__(
        self,
        name: str,
        interval: int = 0,
        look_back: int = 1,
        reset: bool = False,
        stat_logger: CustomLogger = logger,
        log_level: LogLevel = LogLevel.STATS,
    ):
        self.history = deque(maxlen=look_back)
        self.node = None
        self.name = name
        self.log_level = log_level
        self.logger = stat_logger
        self.interval_data = self.get_interval_data_class()()
        self.interval = interval
        self.reset = reset

    @abstractmethod
    def get_interval_data_class(self) -> Type[T]:
        pass

    @abstractmethod
    def get_info(self) -> Optional[Dict[str, Any]]:
        """
        Constructs response object to be outputted on stat service set interval.
        :return: dictionary to be converted to JSON
        """

    def set_node(self, node: N) -> None:
        self.node = node
        # reset the interval data
        self.create_interval_data_object()

    def create_interval_data_object(self) -> None:
        self.interval_data = self.get_interval_data_class()()

    def close_interval_data(self) -> None:
        assert self.node is not None
        assert self.interval_data is not None
        self.interval_data.close()
        self.history.append(self.interval_data)

    # pylint: disable=unused-argument
    def flush_info(self, threshold: int = 0) -> int:
        self.close_interval_data()
        data = self.get_info()
        if data:
            self.logger.log(self.log_level, {"data": data, "type": self.name})

        # Start a new interval data if non cumulative
        if self.reset:
            self.create_interval_data_object()
        return self.interval


class ThreadedStatisticsService(StatisticsService[T, N], metaclass=ABCMeta):
    """
    Abstract class for stats service that may take a long time to execute.
    """

    _thread: Optional[Thread]
    _alive: bool
    _lock: Lock

    def __init__(self, name: str, *args, **kwargs) -> None:
        super(ThreadedStatisticsService, self).__init__(name, *args, **kwargs)
        self._thread = None
        self._alive = True
        self._lock = Lock()

    def start_recording(self, record_fn: Callable) -> None:
        self._thread = Thread(target=self.loop_record_on_thread, args=(record_fn,))
        # pyre-fixme[16]: `Optional` has no attribute `start`.
        self._thread.start()

    def stop_recording(self) -> None:
        # TODO: This is necessary in order to make the tests pass. We are initializing multiple
        # nodes in a process in a test, both of which are initializing the memory_statistics_service.
        # Thus, there is unclear ownership of the global variable. The right fix here is to make
        # memory_statistics_service not a singleton anymore and have it be a variable that is assigned
        # on a per-node basis.
        if self._thread is None:
            self.logger.error(log_messages.STOP_RECORDING_CALLED_ON_UNINITIALIZED_THREAD)
            return

        with self._lock:
            self._alive = False

        # pyre-fixme[16]: `Optional` has no attribute `join`.
        self._thread.join()

    def sleep_and_check_alive(self, sleep_time: float) -> bool:
        """
        Sleeps for sleep_time seconds and checks whether or this service is alive every 30 seconds.
        Returns whether or not this service is alive at the end of this sleep time.
        """

        with self._lock:
            alive = self._alive
        while sleep_time > 0 and alive:
            time.sleep(constants.THREADED_STATS_SLEEP_INTERVAL_S)
            sleep_time -= constants.THREADED_STATS_SLEEP_INTERVAL_S
            with self._lock:
                alive = self._alive

        time.sleep(0)  # ensure sleep is called regardless of the sleep time value
        return alive

    def loop_record_on_thread(self, record_fn: Callable) -> None:
        """
        Assume that record_fn is a read-only function and its okay to get somewhat stale data.
        """
        node = self.node
        assert node is not None

        # align all nodes to write statistics at the same time (clock interval)
        next_clock_interval = ((int(time.time() / self.interval) + 1) * self.interval) - time.time()
        alive = self.sleep_and_check_alive(next_clock_interval)

        while alive:
            start_date_time = datetime.utcnow()
            start_time = time.time()
            try:
                record_fn()
            # pylint: disable=broad-except
            except Exception as e:
                self.logger.error(
                    log_messages.FAILURE_RECORDING_STATS, self.name, e, traceback.format_exc()
                )
            else:
                runtime = time.time() - start_time
                task_duration_logger.statistics(
                    {
                        "type": "TaskDuration",
                        "start_date_time": start_date_time,
                        "task": self.name,
                        "duration": runtime,
                        "node_id": node.opts.node_id,
                    }
                )

            # align all nodes to write statistics at the same time (clock interval)
            next_clock_sleep_time = ((int(time.time() / self.interval) + 1) * self.interval) - time.time()
            alive = self.sleep_and_check_alive(next_clock_sleep_time)



from enum import auto

from bxcommon.models.serializable_flag import SerializableFlag


class ConnectionState(SerializableFlag):
    CONNECTING = auto()
    INITIALIZED = auto()
    HELLO_RECVD = auto()
    HELLO_ACKD = auto()

    # avoid all future usages of ConnectionState.ESTABLISHED
    ESTABLISHED = auto()

    CUT_THROUGH_SOURCE = auto()
    CUT_THROUGH_SINK = auto()
    ULTRA_SLOW = auto()

    def __str__(self) -> str:
        # pylint: disable=using-constant-test
        if self.name:
            return str(self.name)
        else:
            return super().__str__()



import asyncio
from collections import deque
from typing import Optional, List, cast, Iterator, Union

from bxcommon import constants
from bxcommon.connections.abstract_connection import AbstractConnection
from bxcommon.connections.connection_state import ConnectionState
from bxcommon.connections.connection_type import ConnectionType
from bxcommon.feed.feed import FeedKey
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.eth.serializers.transaction import Transaction
from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxcommon.network.ip_endpoint import IpEndpoint
from bxcommon.network.peer_info import ConnectionPeerInfo
from bxcommon.network.transport_layer_protocol import TransportLayerProtocol
from bxcommon.rpc import rpc_constants
from bxcommon.feed.feed_source import FeedSource
from bxcommon.utils import convert
from bxcommon.utils.expiring_dict import ExpiringDict
from bxcommon.utils.blockchain_utils.eth import crypto_utils, eth_common_constants
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats.block_stat_event_type import BlockStatEventType
from bxcommon.utils.stats.block_statistics_service import block_stats
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxgateway import log_messages, gateway_constants
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
from bxgateway.connections.abstract_relay_connection import AbstractRelayConnection
from bxgateway.connections.eth.eth_base_connection_protocol import EthConnectionProtocolStatus
from bxgateway.connections.eth.eth_node_connection import EthNodeConnection
from bxgateway.connections.eth.eth_node_discovery_connection import EthNodeDiscoveryConnection
from bxgateway.connections.eth.eth_relay_connection import EthRelayConnection
from bxgateway.connections.eth.eth_remote_connection import EthRemoteConnection
from bxgateway.connections.gateway_connection import GatewayConnection
from bxgateway.feed.eth.eth_new_block_feed import EthNewBlockFeed
from bxcommon.feed.eth.eth_new_transaction_feed import EthNewTransactionFeed
from bxgateway.feed.eth.eth_on_block_feed import EthOnBlockFeed, EventNotification
from bxcommon.feed.eth.eth_pending_transaction_feed import EthPendingTransactionFeed
from bxgateway.feed.eth.eth_raw_block import EthRawBlock
from bxgateway.feed.eth.eth_transaction_receipts_feed import EthTransactionReceiptsFeed
from bxgateway.messages.eth import eth_message_converter_factory as converter_factory
from bxgateway.messages.eth.internal_eth_block_info import InternalEthBlockInfo
from bxgateway.messages.eth.new_block_parts import NewBlockParts
from bxgateway.messages.eth.protocol.transactions_eth_protocol_message import \
    TransactionsEthProtocolMessage
from bxgateway.rpc.external.eth_ws_proxy_publisher import EthWsProxyPublisher
from bxgateway.services.abstract_block_cleanup_service import AbstractBlockCleanupService
from bxgateway.services.eth.eth_block_processing_service import EthBlockProcessingService
from bxgateway.services.eth.eth_block_queuing_service import EthBlockQueuingService
from bxgateway.services.eth.eth_normal_block_cleanup_service import EthNormalBlockCleanupService
from bxgateway.testing.eth_lossy_relay_connection import EthLossyRelayConnection
from bxgateway.testing.test_modes import TestModes
from bxgateway.utils.interval_minimum import IntervalMinimum
from bxgateway.utils.running_average import RunningAverage
from bxgateway.utils.stats.eth.eth_gateway_stats_service import eth_gateway_stats_service
from bxgateway.utils.stats.eth_on_block_feed_stats_service import eth_on_block_feed_stats_service
from bxutils import logging
from bxutils.services.node_ssl_service import NodeSSLService

logger = logging.get_logger(__name__)


class EthGatewayNode(AbstractGatewayNode):
    DISCONNECT_REASON_TO_DESCRIPTION = {
        0: "Disconnect requested",
        1: "TCP sub-system error",
        2: "Breach of protocol, e.g. a malformed message or bad RLP",
        3: "Useless peer",
        4: "Too many peers",
        5: "Already connected",
        6: "Incompatible P2P protocol version",
        7: "Null node identity received - this is automatically invalid",
        8: "Client quitting",
        9: "Unexpected identity in handshake",
        10: "Identity is the same as this node",
        11: "Ping timeout",
        16: "Some other reason specific to a subprotocol"
    }

    DISCONNECT_REASON_TO_INSTRUCTION = {
        4: "Remove some peers or add gateway as trusted peer.",
        5: "Verify that another instance of the gateway is not already running.",
        16: "Verify that '--blockchain-network' and other blockchain network parameters are correct."
    }

    def __init__(self, opts, node_ssl_service: NodeSSLService) -> None:
        super(EthGatewayNode, self).__init__(opts, node_ssl_service,
                                             eth_common_constants.TRACKED_BLOCK_CLEANUP_INTERVAL_S)

        self.remote_blockchain_protocol_version = eth_common_constants.ETH_PROTOCOL_VERSION
        self.remote_blockchain_connection_established = False

        self._node_public_key = None
        self._remote_public_key = None

        if opts.node_public_key is not None:
            self._node_public_key = convert.hex_to_bytes(opts.node_public_key)
        elif opts.blockchain_peers is None:
            raise RuntimeError(
                "128 digit public key must be included with command-line specified blockchain peer.")
        if opts.remote_blockchain_peer is not None:
            if opts.remote_public_key is None:
                raise RuntimeError(
                    "128 digit public key must be included with command-line specified remote blockchain peer.")
            else:
                self._remote_public_key = convert.hex_to_bytes(opts.remote_public_key)

        self.block_processing_service: EthBlockProcessingService = EthBlockProcessingService(self)
        self.block_parts_storage: ExpiringDict[Sha256Hash, NewBlockParts] = ExpiringDict(
            self.alarm_queue,
            gateway_constants.MAX_BLOCK_CACHE_TIME_S,
            "eth_block_queue_parts",
        )

        # List of know total difficulties, tuples of values (block hash, total difficulty)
        self._last_known_difficulties = deque(maxlen=eth_common_constants.LAST_KNOWN_TOTAL_DIFFICULTIES_MAX_COUNT)

        # queue of the block hashes requested from remote blockchain node during sync
        self._requested_remote_blocks_queue = deque()

        # number of remote block requests to skip in case if requests and responses got out of sync
        self._skip_remote_block_requests_stats_count = 0

        self.init_eth_gateway_stat_logging()
        self.init_eth_on_block_feed_stat_logging()

        self.message_converter = converter_factory.create_eth_message_converter(self.opts)
        self.eth_ws_proxy_publisher = EthWsProxyPublisher(opts.eth_ws_uri, self.feed_manager, self._tx_service, self)
        if self.opts.ws and not self.opts.eth_ws_uri:
            logger.warning(log_messages.ETH_WS_SUBSCRIBER_NOT_STARTED)

        self.average_block_gas_price = RunningAverage(gateway_constants.ETH_GAS_RUNNING_AVERAGE_SIZE)
        self.min_tx_from_node_gas_price = IntervalMinimum(gateway_constants.ETH_MIN_GAS_INTERVAL_S, self.alarm_queue)

        logger.info("Gateway enode url: {}", self.get_enode())

    def build_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        if self._is_in_local_discovery():
            return EthNodeDiscoveryConnection(socket_connection, self)
        else:
            return EthNodeConnection(socket_connection, self)

    def build_relay_connection(self, socket_connection: AbstractSocketConnectionProtocol) -> AbstractRelayConnection:
        if TestModes.DROPPING_TXS in self.opts.test_mode:
            cls = EthLossyRelayConnection
        else:
            cls = EthRelayConnection

        relay_connection = cls(socket_connection, self)
        return relay_connection

    def build_remote_blockchain_connection(
        self, socket_connection: AbstractSocketConnectionProtocol
    ) -> AbstractGatewayBlockchainConnection:
        return EthRemoteConnection(socket_connection, self)

    def build_block_queuing_service(
        self,
        connection: AbstractGatewayBlockchainConnection
    ) -> EthBlockQueuingService:
        return EthBlockQueuingService(self, connection)

    def build_block_cleanup_service(self) -> AbstractBlockCleanupService:
        if self.opts.use_extensions:
            from bxgateway.services.eth.eth_extension_block_cleanup_service import EthExtensionBlockCleanupService
            return EthExtensionBlockCleanupService(self, self.network_num)
        else:
            return EthNormalBlockCleanupService(self, self.network_num)

    def on_updated_remote_blockchain_peer(self, outbound_peer) -> int:
        if "node_public_key" not in outbound_peer.attributes:
            logger.warning(log_messages.BLOCKCHAIN_PEER_LACKS_PUBLIC_KEY)
            return constants.SDN_CONTACT_RETRY_SECONDS
        else:
            super(EthGatewayNode, self).on_updated_remote_blockchain_peer(outbound_peer)
            self._remote_public_key = convert.hex_to_bytes(outbound_peer.attributes["node_public_key"])
            return constants.CANCEL_ALARMS

    def get_outbound_peer_info(self) -> List[ConnectionPeerInfo]:
        peers = []

        for peer in self.opts.outbound_peers:
            peers.append(ConnectionPeerInfo(
                IpEndpoint(peer.ip, peer.port), convert.peer_node_to_connection_type(self.NODE_TYPE, peer.node_type)
            ))

        local_protocol = TransportLayerProtocol.UDP if self._is_in_local_discovery() else TransportLayerProtocol.TCP
        for blockchain_peer in self.blockchain_peers:
            peers.append(ConnectionPeerInfo(
                IpEndpoint(blockchain_peer.ip, blockchain_peer.port),
                ConnectionType.BLOCKCHAIN_NODE,
                local_protocol
            ))

        if self.remote_blockchain_ip is not None and self.remote_blockchain_port is not None:
            remote_protocol = TransportLayerProtocol.UDP if self._is_in_remote_discovery() else \
                TransportLayerProtocol.TCP
            peers.append(ConnectionPeerInfo(
                # pyre-fixme[6]: Expected `str` for 1st param but got `Optional[str]`.
                IpEndpoint(self.remote_blockchain_ip, self.remote_blockchain_port),
                ConnectionType.REMOTE_BLOCKCHAIN_NODE,
                remote_protocol
            ))

        return peers

    def get_private_key(self) -> bytes:
        return convert.hex_to_bytes(self.opts.private_key)

    def get_public_key(self) -> bytes:
        return crypto_utils.private_to_public_key(self.get_private_key())

    def set_node_public_key(self, discovery_connection, node_public_key) -> None:
        if not isinstance(discovery_connection, EthNodeDiscoveryConnection):
            raise TypeError("Argument discovery_connection is expected to be of type EthNodeDiscoveryConnection, was {}"
                            .format(type(discovery_connection)))

        if not node_public_key:
            raise ValueError("node_public_key argument is required")

        self._node_public_key = node_public_key

        # close UDP connection
        discovery_connection.mark_for_close(False)

        # establish TCP connection
        self.enqueue_connection(self.opts.blockchain_ip, self.opts.blockchain_port, ConnectionType.BLOCKCHAIN_NODE)

    def set_remote_public_key(self, discovery_connection, remote_public_key) -> None:
        if not isinstance(discovery_connection, EthNodeDiscoveryConnection):
            raise TypeError("Argument discovery_connection is expected to be of type EthNodeDiscoveryConnection, was {}"
                            .format(type(discovery_connection)))

        if not remote_public_key:
            raise ValueError("remote_public_key argument is required")

        self._remote_public_key = remote_public_key

        # close UDP connection
        discovery_connection.mark_for_close(False)

        remote_blockchain_ip = self.remote_blockchain_ip
        remote_blockchain_port = self.remote_blockchain_port

        assert remote_blockchain_ip is not None and remote_blockchain_port is not None
        # establish TCP connection
        self.enqueue_connection(
            remote_blockchain_ip, remote_blockchain_port, ConnectionType.REMOTE_BLOCKCHAIN_NODE
        )

    def get_node_public_key(self, ip: str, port: int) -> bytes:
        node_public_key = None
        for blockchain_peer in self.blockchain_peers:
            if blockchain_peer.ip == ip and blockchain_peer.port == port:
                node_public_key = blockchain_peer.node_public_key
                break

        if not node_public_key:
            raise RuntimeError(
                f"128 digit public key must be included with for blockchain peer ip {ip} and port {port}."
            )

        return convert.hex_to_bytes(node_public_key)

    def get_remote_public_key(self) -> bytes:
        return self._remote_public_key

    def set_known_total_difficulty(self, block_hash: Sha256Hash, total_difficulty: int) -> None:
        self._last_known_difficulties.append((block_hash, total_difficulty))

    def try_calculate_total_difficulty(self, block_hash: Sha256Hash, new_block_parts: NewBlockParts) -> Optional[int]:
        previous_block_hash = new_block_parts.get_previous_block_hash()
        previous_block_total_difficulty = None

        for known_block_hash, known_total_difficulty in self._last_known_difficulties:
            if previous_block_hash == known_block_hash:
                previous_block_total_difficulty = known_total_difficulty
                break

        if previous_block_total_difficulty is None:
            logger.debug("Unable to calculate total difficulty after block {}.",
                         convert.bytes_to_hex(block_hash.binary))
            return None

        block_total_difficulty = previous_block_total_difficulty + new_block_parts.get_block_difficulty()

        self._last_known_difficulties.append((block_hash, block_total_difficulty))
        logger.debug("Calculated total difficulty after block {} = {}.",
                     convert.bytes_to_hex(block_hash.binary), block_total_difficulty)

        return block_total_difficulty

    def log_requested_remote_blocks(self, block_hashes: List[Sha256Hash]) -> None:
        if self._skip_remote_block_requests_stats_count > 0:
            self._skip_remote_block_requests_stats_count -= 1
        else:
            for block_hash in block_hashes:
                block_stats.add_block_event_by_block_hash(
                    block_hash,
                    BlockStatEventType.REMOTE_BLOCK_REQUESTED_BY_GATEWAY,
                    network_num=self.network_num,
                    more_info=f"Protocol: {self.opts.blockchain_protocol}, "
                    f"Network: {self.opts.blockchain_network}"
                )
            self._requested_remote_blocks_queue.append(block_hashes)

    def log_received_remote_blocks(self, blocks_count: int) -> None:
        if len(self._requested_remote_blocks_queue) > 0:
            expected_blocks = self._requested_remote_blocks_queue.pop()

            if len(expected_blocks) != blocks_count:
                logger.warning(log_messages.BLOCK_COUNT_MISMATCH,
                               blocks_count, len(expected_blocks))
                self._skip_remote_block_requests_stats_count = len(self._requested_remote_blocks_queue) * 2
                self._requested_remote_blocks_queue.clear()
                return

            for block_hash in expected_blocks:
                block_stats.add_block_event_by_block_hash(block_hash,
                                                          BlockStatEventType.REMOTE_BLOCK_RECEIVED_BY_GATEWAY,
                                                          network_num=self.network_num,
                                                          more_info="Protocol: {}, Network: {}".format(
                                                              self.opts.blockchain_protocol,
                                                              self.opts.blockchain_network))
        else:
            logger.warning(log_messages.UNEXPECTED_BLOCKS)

    def _should_decrease_version_number(
        self, connection_state: ConnectionState, connection_status: EthConnectionProtocolStatus
    ) -> bool:
        return (
            ConnectionState.ESTABLISHED not in connection_state
            and connection_status.auth_message_sent
            and not connection_status.auth_message_received
            and not connection_status.auth_ack_message_sent
            and connection_status.auth_ack_message_received
            and connection_status.hello_message_sent
            and connection_status.hello_message_received
            and not connection_status.status_message_sent
            and not connection_status.status_message_received
            and connection_status.disconnect_message_received
            and connection_status.disconnect_reason == 3
        )

    def _get_new_protocol_version(self, connection, peer_version) -> int:
        for index, version in enumerate(eth_common_constants.SUPPORTED_ETH_PROTOCOL_VERSION):
            if (
                version == peer_version
                and index < len(eth_common_constants.SUPPORTED_ETH_PROTOCOL_VERSION) - 1
            ):
                new_version = eth_common_constants.SUPPORTED_ETH_PROTOCOL_VERSION[index + 1]
                connection.log_debug(
                    "Failed to connect with version {}, try to reconnect with version {}",
                    version, new_version
                )
                return new_version

        return peer_version

    def _should_log_closed_connection(self, connection: AbstractConnection) -> bool:
        if isinstance(connection, EthNodeConnection):
            connection_status = connection.connection_protocol.connection_status

            if (
                self._should_decrease_version_number(connection.state, connection_status)
                and connection.CONNECTION_TYPE == ConnectionType.BLOCKCHAIN_NODE
            ):
                for peer in self.blockchain_peers:
                    if (
                        peer == BlockchainPeerInfo(connection.peer_ip, connection.peer_port)
                        and not peer.connection_established
                    ):
                        peer.blockchain_protocol_version = self._get_new_protocol_version(
                            connection, peer.blockchain_protocol_version
                        )
                        return False

        elif isinstance(connection, EthRemoteConnection):
            connection_status = connection.connection_protocol.connection_status

            if (
                self._should_decrease_version_number(connection.state, connection_status)
                and connection.CONNECTION_TYPE == ConnectionType.REMOTE_BLOCKCHAIN_NODE
                and not self.remote_blockchain_connection_established
            ):
                self.remote_blockchain_protocol_version = \
                    self._get_new_protocol_version(
                        connection, self.remote_blockchain_protocol_version
                    )
                return False

        return True

    def log_closed_connection(self, connection: AbstractConnection) -> None:
        if isinstance(connection, EthNodeConnection):
            # pyre-fixme[22]: The cast is redundant.
            eth_node_connection = cast(EthNodeConnection, connection)
            connection_status = connection.connection_protocol.connection_status
            if ConnectionState.INITIALIZED not in eth_node_connection.state:
                logger.info("Failed to connect to Ethereum node. Verify that provided ip address ({}) and port ({}) "
                            "are correct. Verify that firewall port is open. Connection details: {}.",
                            eth_node_connection.peer_ip, eth_node_connection.peer_port, eth_node_connection)
            elif connection_status.disconnect_message_received:
                logger.info("Connection to Ethereum node failed. Disconnect reason: '{}'. {} Connection details: {}.",
                            # pyre-fixme[6]: Expected `int` for 1st param but got
                            #  `Optional[int]`.
                            self._get_disconnect_reason_description(connection_status.disconnect_reason),
                            # pyre-fixme[6]: Expected `int` for 1st param but got
                            #  `Optional[int]`.
                            self._get_disconnect_reason_instruction(connection_status.disconnect_reason),
                            connection)
            elif not connection_status.auth_message_received and not connection_status.auth_ack_message_received:
                logger.info("Failed to connect to Ethereum node. Verify that '--node-public-key' argument is provided "
                            "and value matches enode of the Ethereum node. Connection details: {}.",
                            eth_node_connection)
            elif connection_status.hello_message_received and connection_status.status_message_sent and \
                not connection_status.status_message_received:
                logger.info("Failed to connect to Ethereum node. Verify that '--blockchain-network' and other "
                            "blockchain network arguments are correct. Connection details: {}.",
                            eth_node_connection)
            else:
                super(EthGatewayNode, self).log_closed_connection(connection)

        elif isinstance(connection, GatewayConnection):
            if connection.established:
                logger.debug("Failed to connect to: {}.", connection)
            else:
                logger.debug("Closed connection: {}", connection)
        else:
            super(EthGatewayNode, self).log_closed_connection(connection)

    def init_eth_gateway_stat_logging(self) -> None:
        eth_gateway_stats_service.set_node(self)
        self.alarm_queue.register_alarm(eth_gateway_stats_service.interval, eth_gateway_stats_service.flush_info)

    def init_eth_on_block_feed_stat_logging(self) -> None:
        eth_on_block_feed_stats_service.set_node(self)
        self.alarm_queue.register_alarm(eth_on_block_feed_stats_service.interval,
                                        eth_on_block_feed_stats_service.flush_info)

    def init_live_feeds(self) -> None:
        self.feed_manager.register_feed(EthNewTransactionFeed(network_num=self.network_num))
        self.feed_manager.register_feed(EthPendingTransactionFeed(self.alarm_queue, network_num=self.network_num))
        self.feed_manager.register_feed(EthOnBlockFeed(self, network_num=self.network_num))
        self.feed_manager.register_feed(EthNewBlockFeed(self, network_num=self.network_num))
        self.feed_manager.register_feed(EthTransactionReceiptsFeed(self, network_num=self.network_num))

    def on_new_subscriber_request(self) -> None:
        if self.opts.eth_ws_uri and not self.eth_ws_proxy_publisher.running:
            asyncio.create_task(self.eth_ws_proxy_publisher.revive())

    def on_transactions_in_block(self, transactions: List[Transaction]) -> None:
        for transaction in transactions:
            self.average_block_gas_price.add_value(transaction.adjusted_gas_price())

    def broadcast_transactions_to_nodes(
        self, msg: AbstractMessage, broadcasting_conn: Optional[AbstractConnection]
    ) -> bool:
        msg = cast(TransactionsEthProtocolMessage, msg)

        if self.opts.filter_txs_factor > 0:
            average_block_gas_filter = self.average_block_gas_price.average * self.opts.filter_txs_factor
        else:
            average_block_gas_filter = 0
        min_gas_price_from_node = self.min_tx_from_node_gas_price.current_minimum

        gas_price_filter = max(average_block_gas_filter, min_gas_price_from_node)

        if gas_price_filter > 0:
            assert len(msg.get_transactions()) == 1
            transaction = msg.get_transactions()[0]

            gas_price = float(transaction.adjusted_gas_price())

            if gas_price < gas_price_filter:
                logger.trace(
                    "Skipping sending transaction {} with gas price: {}. Average was {}. Minimum from node was {}.",
                    transaction.hash(),
                    gas_price,
                    average_block_gas_filter,
                    min_gas_price_from_node
                )
                tx_stats.add_tx_by_hash_event(
                    transaction.hash(),
                    TransactionStatEventType.TX_FROM_BDN_IGNORE_LOW_GAS_PRICE,
                    self.network_num,
                    peers=[broadcasting_conn],
                    more_info="Tx gas price {}. Average block gas price: {}. Node min gas price {}."
                        .format(gas_price, average_block_gas_filter, min_gas_price_from_node)
                )
                return False

        return super().broadcast_transactions_to_nodes(msg, broadcasting_conn)

    def get_enode(self) -> str:
        return \
            f"enode://{convert.bytes_to_hex(self.get_public_key())}@{self.opts.external_ip}:{self.opts.non_ssl_port}"

    def on_block_received_from_bdn(
        self, block_hash: Sha256Hash, block_message: AbstractBlockMessage
    ) -> None:
        super().on_block_received_from_bdn(block_hash, block_message)

        node_conn = cast(EthNodeConnection, self.get_any_active_blockchain_connection())
        if node_conn is not None:
            eth_connection_protocol = node_conn.connection_protocol
            eth_connection_protocol.pending_new_block_parts.remove_item(block_hash)

    # pyre-fixme[14]: Inconsistent override:
    # Parameter of type InternalEthBlockInfo is not a supertype of AbstractBlockMessage
    def publish_block(
        self,
        block_number: Optional[int],
        block_hash: Sha256Hash,
        block_message: Optional[InternalEthBlockInfo],
        source: FeedSource
    ) -> None:
        if block_number is None and block_message is not None:
            block_number = block_message.block_number()
        logger.debug(
            "Handle block notification for feed. Number: {}, Hash: {} Msg: {} From: {}",
            block_number, block_hash, block_message, source
        )
        if block_number and self.opts.ws:
            raw_block = EthRawBlock(
                block_number,
                block_hash,
                source,
                self._get_block_message_lazy(block_message, block_hash)
            )
            self._publish_block_to_on_block_feed(raw_block)
            self._publish_block_to_transaction_receipts_feed(raw_block)
            if block_message is not None:
                self._publish_block_to_new_block_feed(raw_block)

    async def init(self) -> None:
        await super().init()
        try:
            await asyncio.wait_for(
                self.eth_ws_proxy_publisher.start(), rpc_constants.RPC_SERVER_INIT_TIMEOUT_S
            )
        except Exception as e:
            logger.error(log_messages.ETH_WS_INITIALIZATION_FAIL, e)
            self.should_force_exit = True

    async def close(self) -> None:
        try:
            await asyncio.wait_for(
                self.eth_ws_proxy_publisher.stop(),
                rpc_constants.RPC_SERVER_STOP_TIMEOUT_S
            )
        except Exception as e:
            logger.error(log_messages.ETH_WS_CLOSE_FAIL, e, exc_info=True)
        await super().close()

    def _is_in_local_discovery(self) -> bool:
        return not self.opts.no_discovery and self._node_public_key is None

    def _is_in_remote_discovery(self) -> bool:
        return not self.opts.no_discovery and self._remote_public_key is None

    def _get_disconnect_reason_description(self, reason: int) -> str:
        if reason not in self.DISCONNECT_REASON_TO_DESCRIPTION:
            return f"Disconnect reason ({reason}) is unknown."

        return self.DISCONNECT_REASON_TO_DESCRIPTION[reason]

    def _get_disconnect_reason_instruction(self, reason: int) -> str:
        if reason not in self.DISCONNECT_REASON_TO_INSTRUCTION:
            return ""

        return self.DISCONNECT_REASON_TO_INSTRUCTION[reason]

    def _get_block_message_lazy(
        self, block_message: Optional[InternalEthBlockInfo], block_hash
    ) -> Iterator[InternalEthBlockInfo]:
        if block_message:
            yield block_message
        else:
            block_queuing_service = cast(
                EthBlockQueuingService,
                self.block_queuing_service_manager.get_designated_block_queuing_service()
            )
            if block_queuing_service is not None:
                block_parts = block_queuing_service.get_block_parts(block_hash)
                if block_parts:
                    block_message = InternalEthBlockInfo.from_new_block_parts(block_parts)
                    yield block_message

    def _publish_block_to_new_block_feed(
        self,
        raw_block: EthRawBlock,
    ) -> None:
        self.feed_manager.publish_to_feed(
            FeedKey(EthNewBlockFeed.NAME, network_num=self.network_num),
            raw_block
        )

    def _publish_block_to_on_block_feed(
        self,
        raw_block: EthRawBlock,
    ) -> None:
        if raw_block.source in EthOnBlockFeed.VALID_SOURCES:
            self.feed_manager.publish_to_feed(
                FeedKey(EthOnBlockFeed.NAME, network_num=self.network_num),
                EventNotification(raw_block.block_number)
            )

    def _publish_block_to_transaction_receipts_feed(
        self,
        raw_block: EthRawBlock
    ):
        self.feed_manager.publish_to_feed(
            FeedKey(EthTransactionReceiptsFeed.NAME, network_num=self.network_num),
            raw_block
        )

    def is_gas_price_above_min_network_fee(self, transaction_contents: Union[memoryview, bytearray]) -> bool:
        # gas_price = eth_common_utils.raw_tx_gas_price(memoryview(transaction_contents), 0)
        # if gas_price >= self.get_blockchain_network().min_tx_network_fee:
        return True
        # return False



import asyncio
import json
from typing import TYPE_CHECKING, Dict, Any, Optional, Union, cast, Type, Tuple

from bxcommon.feed.feed import FeedKey
from bxcommon.rpc.abstract_rpc_handler import AbstractRpcHandler
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.rpc.requests.transaction_status_rpc_request import TransactionStatusRpcRequest
from bxcommon.rpc.rpc_request_type import RpcRequestType
from bxcommon.feed.feed_manager import FeedManager
from bxcommon.feed.subscriber import Subscriber
from bxcommon.rpc.abstract_ws_rpc_handler import Subscription

from bxgateway import gateway_constants, log_messages
from bxgateway.rpc.requests.add_blockchain_peer_rpc_request import AddBlockchainPeerRpcRequest
from bxgateway.rpc.requests.bdn_performance_rpc_request import BdnPerformanceRpcRequest
from bxgateway.rpc.requests.gateway_blxr_transaction_rpc_request import \
    GatewayBlxrTransactionRpcRequest
from bxgateway.rpc.requests.gateway_memory_rpc_request import GatewayMemoryRpcRequest
from bxgateway.rpc.requests.gateway_memory_usage_report_rpc_request import GatewayMemoryUsageRpcRequest
from bxgateway.rpc.requests.gateway_peers_rpc_request import GatewayPeersRpcRequest
from bxgateway.rpc.requests.gateway_status_rpc_request import GatewayStatusRpcRequest
from bxgateway.rpc.requests.gateway_stop_rpc_request import GatewayStopRpcRequest
from bxgateway.rpc.requests.gateway_subscribe_rpc_request import GatewaySubscribeRpcRequest
from bxgateway.rpc.requests.gateway_transaction_service_rpc_request import GatewayTransactionServiceRpcRequest
from bxgateway.rpc.requests.quota_usage_rpc_request import QuotaUsageRpcRequest
from bxgateway.rpc.requests.remove_blockchain_peer_rpc_request import RemoveBlockchainPeerRpcRequest
from bxcommon.rpc.requests.subscribe_rpc_request import SubscribeRpcRequest
from bxcommon.rpc.requests.unsubscribe_rpc_request import UnsubscribeRpcRequest
from bxgateway.rpc.requests.gateway_blxr_call_rpc_request import GatewayBlxrCallRpcRequest
from bxutils import logging
from bxutils.encoding.json_encoder import Case

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode


logger = logging.get_logger(__name__)


class SubscriptionRpcHandler(AbstractRpcHandler["AbstractGatewayNode", Union[bytes, str], Union[bytes, str]]):
    feed_manager: FeedManager
    subscriptions: Dict[str, Subscription]
    subscribed_messages: 'asyncio.Queue[BxJsonRpcRequest]'

    def __init__(self, node: "AbstractGatewayNode", feed_manager: FeedManager, case: Case) -> None:
        super().__init__(node, case)
        self.request_handlers = {
            RpcRequestType.BLXR_TX: GatewayBlxrTransactionRpcRequest,
            RpcRequestType.BLXR_ETH_CALL: GatewayBlxrCallRpcRequest,
            RpcRequestType.GATEWAY_STATUS: GatewayStatusRpcRequest,
            RpcRequestType.STOP: GatewayStopRpcRequest,
            RpcRequestType.MEMORY: GatewayMemoryRpcRequest,
            RpcRequestType.PEERS: GatewayPeersRpcRequest,
            RpcRequestType.BDN_PERFORMANCE: BdnPerformanceRpcRequest,
            RpcRequestType.SUBSCRIBE: GatewaySubscribeRpcRequest,
            RpcRequestType.UNSUBSCRIBE: UnsubscribeRpcRequest,
            RpcRequestType.QUOTA_USAGE: QuotaUsageRpcRequest,
            RpcRequestType.MEMORY_USAGE: GatewayMemoryUsageRpcRequest,
            RpcRequestType.TX_STATUS: TransactionStatusRpcRequest,
            RpcRequestType.TX_SERVICE: GatewayTransactionServiceRpcRequest,
            RpcRequestType.ADD_BLOCKCHAIN_PEER: AddBlockchainPeerRpcRequest,
            RpcRequestType.REMOVE_BLOCKCHAIN_PEER: RemoveBlockchainPeerRpcRequest,
        }

        self.feed_manager = feed_manager
        self.subscriptions = {}
        self.subscribed_messages = asyncio.Queue(
            gateway_constants.RPC_SUBSCRIBER_MAX_QUEUE_SIZE
        )
        self.disconnect_event = asyncio.Event()

    async def parse_request(self, request: Union[bytes, str]) -> Dict[str, Any]:
        return json.loads(request)

    def get_request_handler(self, request: BxJsonRpcRequest) -> AbstractRpcRequest:
        if request.method == RpcRequestType.SUBSCRIBE:
            return self._subscribe_request_factory(request)
        elif request.method == RpcRequestType.UNSUBSCRIBE:
            return self._unsubscribe_request_factory(request)
        else:
            request_handler_type = self.request_handlers[request.method]
            return request_handler_type(request, self.node)

    def serialize_response(self, response: JsonRpcResponse) -> str:
        return response.to_jsons(self.case)

    def serialize_cached_subscription_message(self, message: BxJsonRpcRequest) -> bytes:
        return self.node.serialized_message_cache.serialize_from_cache(message, self.case)

    async def get_next_subscribed_message(self) -> BxJsonRpcRequest:
        return await self.subscribed_messages.get()

    async def handle_subscription(self, subscriber: Subscriber) -> None:
        while True:
            notification = await subscriber.receive()
            # subscription notifications are sent as JSONRPC requests
            next_message = BxJsonRpcRequest(
                None,
                RpcRequestType.SUBSCRIBE,
                {
                    "subscription": subscriber.subscription_id,
                    "result": notification
                }
            )
            if self.subscribed_messages.full():
                logger.info(
                    log_messages.GATEWAY_BAD_FEED_SUBSCRIBER,
                    list(self.subscriptions.keys()),
                    self.subscribed_messages.qsize()
                )
                asyncio.create_task(self.async_close())
                return
            else:
                await self.subscribed_messages.put(next_message)

    async def wait_for_close(self) -> None:
        await self.disconnect_event.wait()

    async def async_close(self) -> None:
        self.close()

    def close(self) -> None:
        subscription_ids = list(self.subscriptions.keys())
        for subscription_id in subscription_ids:
            feed_key, account_id = self._on_unsubscribe(subscription_id)
            assert feed_key is not None
            self.feed_manager.unsubscribe_from_feed(
                feed_key,
                subscription_id,
                account_id
            )
        self.subscriptions = {}

        self.disconnect_event.set()

    def _on_new_subscriber(self, subscriber: Subscriber, feed_key: FeedKey, account_id: Optional[str] = None) -> None:
        task = asyncio.ensure_future(self.handle_subscription(subscriber))
        self.subscriptions[subscriber.subscription_id] = Subscription(
            subscriber, feed_key, task, account_id
        )
        self.node.on_new_subscriber_request()

    def _on_unsubscribe(self, subscriber_id: str) -> Tuple[Optional[FeedKey], Optional[str]]:
        if subscriber_id in self.subscriptions:
            (subscriber, feed_key, task, account_id) = self.subscriptions.pop(subscriber_id)
            task.cancel()
            return feed_key, account_id
        return None, None

    def _subscribe_request_factory(
        self, request: BxJsonRpcRequest
    ) -> AbstractRpcRequest:
        subscribe_rpc_request = cast(Type[SubscribeRpcRequest], self.request_handlers[RpcRequestType.SUBSCRIBE])
        return subscribe_rpc_request(
            request, self.node, self.feed_manager, self._on_new_subscriber
        )

    def _unsubscribe_request_factory(
        self, request: BxJsonRpcRequest
    ) -> AbstractRpcRequest:
        unsubscribe_rpc_request = cast(Type[UnsubscribeRpcRequest], self.request_handlers[RpcRequestType.UNSUBSCRIBE])
        return unsubscribe_rpc_request(
            request, self.node, self.feed_manager, self._on_unsubscribe
        )



import blxr_rlp as rlp
from blxr_rlp.sedes import CountableList

from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType


class DisconnectEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.DISCONNECT

    fields = [("reason", CountableList(rlp.sedes.big_endian_int))]

    def get_reason(self):
        return self.get_field_value("reason")[0]



import json
from typing import List, Any, Union, Dict, Optional, Tuple

import humps
import orjson

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.rpc_errors import RpcMethodNotFound
from bxutils.encoding import json_encoder
from bxutils.encoding.json_encoder import Case


class JsonRpcRequest:
    id: Optional[str]
    method_name: str
    params: Union[List[Any], Dict[Any, Any], None]

    def __init__(
        self,
        request_id: Optional[str],
        method: str,
        params: Union[List[Any], Dict[Any, Any], None],
    ) -> None:
        self.id = request_id
        self.method_name = method
        self.params = params
        self.json_rpc_version = rpc_constants.JSON_RPC_VERSION

    def __str__(self) -> str:
        return f"JsonRpcRequest<{self.to_json()}>"

    def __eq__(self, o: object) -> bool:
        if isinstance(o, JsonRpcRequest):
            return (
                self.id == o.id
                and self.method_name == o.method_name
                and self.params == o.params
                and self.json_rpc_version == o.json_rpc_version
            )
        else:
            return False

    def to_json(self, case: Case = Case.SNAKE) -> Dict[str, Any]:
        params = self.params
        if params is not None and case == Case.CAMEL:
            params = humps.camelize(params)

        return {
            rpc_constants.JSON_RPC_VERSION_FIELD: self.json_rpc_version,
            rpc_constants.JSON_RPC_REQUEST_ID: self.id,
            rpc_constants.JSON_RPC_METHOD: self.method_name,
            rpc_constants.JSON_RPC_PARAMS: params,
        }

    def to_jsons(self, case: Case = Case.SNAKE) -> str:
        json_dict = self.to_json(case)
        return json_encoder.to_json(json_dict)

    def to_json_bytes(self, case: Case = Case.SNAKE) -> bytes:
        json_dict = self.to_json(case)
        return orjson.dumps(json_dict)

    def to_json_bytes_split_serialization(self, case: Case) -> Tuple[bytes, str]:
        """
        This method should only be called for subscription messages.

        No runtime checks are added for performance reasons, so please make sure
        that other RPC request types don't trigger this.

        Returns result serialized separately for caching later.
        """
        json_dict = self.to_json(case)
        result = orjson.dumps(json_dict[rpc_constants.JSON_RPC_PARAMS]["result"]).decode("utf-8")
        json_dict[rpc_constants.JSON_RPC_PARAMS]["result"] = result
        return orjson.dumps(json_dict), result

    def to_json_bytes_with_cached_result(self, case: Case, cached_result: str) -> bytes:
        # noinspection PyTypeChecker
        params = self.params
        assert isinstance(params, dict)
        params["result"] = cached_result
        return self.to_json_bytes(case)

    @classmethod
    def from_json(cls, payload: Dict[str, Any]) -> "JsonRpcRequest":
        request_id = payload.get(rpc_constants.JSON_RPC_REQUEST_ID, None)
        if rpc_constants.JSON_RPC_METHOD not in payload:
            raise RpcMethodNotFound(request_id, "RPC request does not contain a method.")

        method = payload[rpc_constants.JSON_RPC_METHOD]
        request_params = payload.get(rpc_constants.JSON_RPC_PARAMS, None)
        return cls(request_id, method, request_params)

    @classmethod
    def from_jsons(cls, payload: Union[bytes, str]) -> "JsonRpcRequest":
        return cls.from_json(json.loads(payload))



import weakref
from typing import TYPE_CHECKING

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.network.abstract_socket_connection_protocol import AbstractSocketConnectionProtocol
from bxgateway.connections.abstract_gateway_blockchain_connection import AbstractGatewayBlockchainConnection
from bxgateway.connections.ont.ont_node_connection_protocol import OntNodeConnectionProtocol
from bxgateway.messages.ont.ont_message_factory import ont_message_factory
from bxgateway.messages.ont.ping_ont_message import PingOntMessage

if TYPE_CHECKING:
    from bxgateway.connections.ont.ont_gateway_node import OntGatewayNode


class OntNodeConnection(AbstractGatewayBlockchainConnection["OntGatewayNode"]):
    """
    bloXroute gateway <=> blockchain node connection class.
    """

    def __init__(self, sock: AbstractSocketConnectionProtocol, node: "OntGatewayNode"):
        super(OntNodeConnection, self).__init__(sock, node)
        self.connection_protocol = weakref.ref(OntNodeConnectionProtocol(self))

    def connection_message_factory(self) -> AbstractMessageFactory:
        return ont_message_factory

    def ping_message(self) -> AbstractMessage:
        return PingOntMessage(
            self.node.opts.blockchain_net_magic, height=self.node.current_block_height
        )



from typing import List, Dict, Any

import blxr_rlp as rlp

from bxcommon.messages.eth.serializers.block_header import BlockHeader
from bxcommon.messages.eth.serializers.transaction import Transaction


# pyre-fixme[13]: Attribute `header` is never initialized.
class Block(rlp.Serializable):
    fields = [
        ("header", BlockHeader),
        ("transactions", rlp.sedes.CountableList(Transaction)),
        ("uncles", rlp.sedes.CountableList(BlockHeader))
    ]

    header: BlockHeader
    transactions: List[Transaction]
    uncles: List[BlockHeader]

    def __init__(self, *args, **kwargs):
        self.transactions = []
        self.uncles = []

        super().__init__(*args, **kwargs)

    def to_json(self) -> Dict[str, Any]:
        """
        Serializes data for publishing to the block feed.
        """
        block_json = {
            "header": self.header.to_json(),
            "transactions": [],
            "uncles": []
        }
        transactions = []
        for transaction in self.transactions:
            transactions.append(transaction.to_json())
        block_json["transactions"] = transactions

        uncles = []
        for uncle in self.uncles:
            uncles.append(uncle.to_json())
        block_json["uncles"] = uncles

        return block_json



import time
from collections import deque
from datetime import datetime
from typing import Tuple, Optional, Set

from bxcommon.utils import convert
from bxcommon.utils.proxy import task_pool_proxy
from bxcommon.services.extension_transaction_service import ExtensionTransactionService
from bxcommon.utils import memory_utils
from bxcommon.utils.blockchain_utils.ont.ont_object_hash import OntObjectHash
from bxcommon.utils.memory_utils import SpecialTuple
from bxcommon.utils.proxy.task_queue_proxy import TaskQueueProxy
from bxcommon.utils.object_hash import Sha256Hash

from bxgateway import ont_constants
from bxgateway.abstract_message_converter import BlockDecompressionResult
from bxgateway.messages.ont import ont_normal_message_converter
from bxgateway.messages.ont.abstract_ont_message_converter import AbstractOntMessageConverter, get_block_info
from bxgateway.messages.ont.block_ont_message import BlockOntMessage
from bxgateway.utils.block_info import BlockInfo
from bxgateway.utils.errors import message_conversion_error

from bxutils import logging


import task_pool_executor as tpe

logger = logging.get_logger(__name__)


class OntExtensionMessageConverter(AbstractOntMessageConverter):

    DEFAULT_BLOCK_SIZE = ont_constants.ONT_DEFAULT_BLOCK_SIZE
    MINIMAL_SUB_TASK_TX_COUNT = ont_constants.ONT_MINIMAL_SUB_TASK_TX_COUNT

    def __init__(self, ont_magic: int):
        super(OntExtensionMessageConverter, self).__init__(ont_magic)
        self._default_block_size = self.DEFAULT_BLOCK_SIZE
        self.compression_tasks = TaskQueueProxy(self._create_compression_task)
        self.decompression_tasks = TaskQueueProxy(self._create_decompression_task)

    def block_to_bx_block(
        self, block_msg, tx_service, enable_block_compression: bool, min_tx_age_seconds: float
    ) -> Tuple[memoryview, BlockInfo]:
        compress_start_datetime = datetime.utcnow()
        compress_start_timestamp = time.time()
        self._default_block_size = max(self._default_block_size, len(block_msg.buf))
        tsk = self.compression_tasks.borrow_task()
        tsk.init(tpe.InputBytes(block_msg.buf), tx_service.proxy, enable_block_compression, min_tx_age_seconds)
        try:
            task_pool_proxy.run_task(tsk)
        except tpe.AggregatedException as e:
            self.compression_tasks.return_task(tsk)
            raise message_conversion_error.btc_block_compression_error(block_msg.block_hash(), e)
        bx_block = tsk.bx_block()
        block = memoryview(bx_block)
        compressed_size = len(block)
        original_size = len(block_msg.rawbytes())
        block_hash = OntObjectHash(binary=convert.hex_to_bytes(tsk.block_hash().hex_string()))

        block_info = BlockInfo(
            block_hash,
            tsk.short_ids(),
            compress_start_datetime,
            datetime.utcnow(),
            (time.time() - compress_start_timestamp) * 1000,
            tsk.txn_count(),
            tsk.compressed_block_hash().hex_string(),
            tsk.prev_block_hash().hex_string(),
            original_size,
            compressed_size,
            100 - float(compressed_size) / original_size * 100,
            tsk.ignored_short_ids()
        )
        self.compression_tasks.return_task(tsk)
        return block, block_info

    def bx_block_to_block(
            self, bx_block_msg: memoryview, tx_service: ExtensionTransactionService
    ) -> BlockDecompressionResult:
        decompress_start_datetime = datetime.utcnow()
        decompress_start_timestamp = time.time()
        tsk = self.decompression_tasks.borrow_task()
        tsk.init(tpe.InputBytes(bx_block_msg), tx_service.proxy)
        try:
            task_pool_proxy.run_task(tsk)
        except tpe.AggregatedException as e:
            self.decompression_tasks.return_task(tsk)
            header_info = ont_normal_message_converter.parse_bx_block_header(bx_block_msg, deque())
            raise message_conversion_error.btc_block_decompression_error(header_info.block_hash, e)
        total_tx_count = tsk.tx_count()
        unknown_tx_hashes = [Sha256Hash(bytearray(unknown_tx_hash.binary()))
                             for unknown_tx_hash in tsk.unknown_tx_hashes()]
        unknown_tx_sids = tsk.unknown_tx_sids()
        block_hash = OntObjectHash(
            binary=convert.hex_to_bytes(tsk.block_hash().hex_string())
        )
        if tsk.success():
            ont_block_msg = BlockOntMessage(buf=memoryview(tsk.block_message()))
            logger.debug(
                "Successfully parsed block broadcast message. {} transactions "
                "in block {}",
                total_tx_count,
                block_hash
            )
        else:
            ont_block_msg = None
            logger.debug(
                "Block recovery needed for {}. Missing {} sids, {} tx hashes. "
                "Total txs in block: {}",
                block_hash,
                len(unknown_tx_sids),
                len(unknown_tx_hashes),
                total_tx_count
            )
        block_info = get_block_info(
            bx_block_msg,
            block_hash,
            tsk.short_ids(),
            decompress_start_datetime,
            decompress_start_timestamp,
            total_tx_count,
            ont_block_msg
        )
        self.decompression_tasks.return_task(tsk)
        return BlockDecompressionResult(ont_block_msg, block_info, unknown_tx_sids, unknown_tx_hashes)

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return memory_utils.add_special_objects(self.compression_tasks, self.decompression_tasks, ids=ids)

    def _create_compression_task(self) -> tpe.OntBlockCompressionTask:
        return tpe.OntBlockCompressionTask(self._default_block_size, self.MINIMAL_SUB_TASK_TX_COUNT)

    def _create_decompression_task(self) -> tpe.OntBlockDecompressionTask:
        return tpe.OntBlockDecompressionTask(self._default_block_size, self.MINIMAL_SUB_TASK_TX_COUNT)



GENERAL_CATEGORY = "General category"
PROCESSING_FAILED_CATEGORY = "Processing problem"
LOGGING_CATEGORY = "Logging problem"
MEMORY_CATEGORY = "Memory problem"
REQUEST_RESPONSE_CATEGORY = "Unsatisfactory data in the response"
NETWORK_CATEGORY = "Networking problem"
CONNECTION_PROBLEM_CATEGORY = "Connection problem"
THREAD_CATEGORY = "Threading problems"
UNCATEGORIZED = "Uncategorized"
THIRD_PARTY_CATEGORY = "Third party log message"
FAILED_TO_ACQUIRE_REDIS_LOCK = "Failed to acquire redis lock"
NETWORK_ERROR = "Network error"
NODE_ID_ERROR = "Missing or incorrect node id"
MISSING_NODE = "Missing node"
EXCEPTION_OR_CANCELLATION = "Exception error or task cancelled"
EVENT_ERROR = "Event missing, incorrect or incompatible"
CONFIGURATION_ERROR = "Error setting up or reading configuration"
GENERAL_WARNING = "General warning category"
AUTHENTICATION_ERROR = "Authentication error"
RPC_ERROR = "Rpc error"
AUDITING_ERROR = "Auditing error"
NOTIFICATION_FROM_RELAY_CATEGORY = "Notification from relay"



# TODO: these functions arent intended to be permanent, but pulling
# them out here for now for easier tests to convert these to different
# formats for symmetric encryption / hashing.
# We can probably implement some sort of EncryptionCacheProtocol that
# converts hash/payload types to avoid the copies on block_id (and so we
# can make ObjectHash the storage key), but that doesn't solve the copy
# of the payload, which is probably the mort significant part.
from bxcommon import constants
from bxcommon.storage.encrypted_cache import EncryptedCache
from bxcommon.utils.object_hash import Sha256Hash


def message_hash_to_hash_key(msg_hash) -> bytes:
    if isinstance(msg_hash, Sha256Hash):
        return bytes(msg_hash.binary)

    if isinstance(msg_hash, memoryview):
        return msg_hash.tobytes()

    return bytes(msg_hash)


def message_blob_to_ciphertext(msg_blob) -> bytes:
    return msg_blob.tobytes()


class BlockEncryptedCache(EncryptedCache):
    def __init__(self, alarm_queue) -> None:
        super(BlockEncryptedCache, self).__init__(constants.BLOCK_CACHE_TIMEOUT_S, alarm_queue)

    def decrypt_ciphertext(self, hash_key, ciphertext):
        """
        Attempts to decrypt from hash and blob from a BroadcastMessage
        :param hash_key: BroadcastMessage.block_id()
        :param ciphertext: BroadcastMessage.blob()
        :return decrypted block
        """
        return super(BlockEncryptedCache, self).decrypt_ciphertext(
            message_hash_to_hash_key(hash_key),
            message_blob_to_ciphertext(ciphertext)
        )

    def decrypt_and_get_payload(self, hash_key, encryption_key):
        """
        Attempts to decrypt from hash and key from a KeyMessage
        :param hash_key: KeyMessage.hash_key()
        :param encryption_key: KeyMessage.key()
        :return:  decrypted block
        """
        return super(BlockEncryptedCache, self).decrypt_and_get_payload(
            message_hash_to_hash_key(hash_key),
            message_blob_to_ciphertext(encryption_key)
        )

    def has_encryption_key_for_hash(self, hash_key):
        return super(BlockEncryptedCache, self).has_encryption_key_for_hash(
            message_hash_to_hash_key(hash_key),
        )

    def has_ciphertext_for_hash(self, hash_key):
        return super(BlockEncryptedCache, self).has_ciphertext_for_hash(
            message_hash_to_hash_key(hash_key),
        )

    def add_ciphertext(self, hash_key, ciphertext):
        return super(BlockEncryptedCache, self).add_ciphertext(
            message_hash_to_hash_key(hash_key),
            message_blob_to_ciphertext(ciphertext)
        )

    def add_key(self, hash_key, encryption_key):
        return super(BlockEncryptedCache, self).add_key(
            message_hash_to_hash_key(hash_key),
            message_blob_to_ciphertext(encryption_key)
        )

    def pop_ciphertext(self, hash_key):
        return super(BlockEncryptedCache, self).pop_ciphertext(
            message_hash_to_hash_key(hash_key)
        )

    def remove_item(self, hash_key):
        return super(BlockEncryptedCache, self).remove_item(
            message_hash_to_hash_key(hash_key)
        )



from typing import List

import blxr_rlp as rlp
from bxcommon.utils.object_hash import Sha256Hash
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxgateway.messages.eth.serializers.transient_block_body import TransientBlockBody
from bxcommon.utils.blockchain_utils.eth import rlp_utils, eth_common_utils
from bxutils.logging.log_level import LogLevel


class BlockBodiesEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.BLOCK_BODIES

    fields = [("blocks", rlp.sedes.CountableList(TransientBlockBody))]

    def __init__(self, msg_bytes, *args, **kwargs):
        super(BlockBodiesEthProtocolMessage, self).__init__(msg_bytes, *args, **kwargs)

        self._block_body_bytes = None

    def __repr__(self):
        return f"BlockBodiesEthProtocolMessage<bodies_count: {len(self.get_block_bodies_bytes())}>"

    def get_blocks(self) -> List[TransientBlockBody]:
        return self.get_field_value("blocks")

    def get_block_bodies_bytes(self):
        if self._block_body_bytes:
            return self._block_body_bytes

        if self._memory_view is None:
            self.serialize()

        self._block_body_bytes = rlp_utils.get_first_list_field_items_bytes(self._memory_view)

        return self._block_body_bytes

    def get_block_transaction_hashes(self, block_index: int) -> List[Sha256Hash]:
        tx_hashes = []

        for tx_bytes in self.get_block_transaction_bytes(block_index):
            tx_hash_bytes = eth_common_utils.keccak_hash(tx_bytes)
            tx_hashes.append(Sha256Hash(tx_hash_bytes))

        return tx_hashes

    def get_block_transaction_bytes(self, block_index: int) -> List[memoryview]:
        block_body_bytes = self.get_block_bodies_bytes()

        if block_index >= len(block_body_bytes):
            return []

        txs_bytes = rlp_utils.get_first_list_field_items_bytes(block_body_bytes[block_index])[0]

        return rlp_utils.get_first_list_field_items_bytes(txs_bytes)

    @classmethod
    def from_body_bytes(cls, body_bytes: memoryview) -> "BlockBodiesEthProtocolMessage":
        bodies_list_prefix = rlp_utils.get_length_prefix_list(len(body_bytes))

        msg_bytes = bytearray(len(bodies_list_prefix) + len(body_bytes))
        msg_bytes[:len(bodies_list_prefix)] = bodies_list_prefix
        msg_bytes[len(bodies_list_prefix):] = body_bytes

        return cls(msg_bytes)

    def log_level(self):
        return LogLevel.DEBUG



from bxcommon import constants
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc import rpc_constants
from bxgateway.rpc.requests.abstract_blockchain_peer_rpc_request import AbstractBlockchainPeerRpcRequest
from bxgateway.utils.logging.status import status_log


class RemoveBlockchainPeerRpcRequest(AbstractBlockchainPeerRpcRequest):
    help = {
        "params": f"{rpc_constants.BLOCKCHAIN_PEER_PARAMS_KEY}: "
                "For Ethereum, the format is enode://<eth node public key>@<eth node ip>:<port>. "
                "For other blockchain protocols, the format is <ip>:<port>",
        "description": "Remove blockchain peer"
    }

    async def process_request(self) -> JsonRpcResponse:
        blockchain_peer_info = self._blockchain_peer_info
        assert blockchain_peer_info is not None
        self.node.blockchain_peers.discard(blockchain_peer_info)
        if self.node.connection_pool.has_connection(blockchain_peer_info.ip, blockchain_peer_info.port):
            peer_conn_to_remove = self.node.connection_pool.get_by_ipport(
                blockchain_peer_info.ip, blockchain_peer_info.port
            )
            peer_conn_to_remove.mark_for_close(False)
        else:
            self.node.time_blockchain_peer_conn_destroyed_by_ip.pop(
                (blockchain_peer_info.ip, blockchain_peer_info.port), None
            )
            self.node.alarm_queue.register_approx_alarm(
                2 * constants.MIN_SLEEP_TIMEOUT, constants.MIN_SLEEP_TIMEOUT, status_log.update_alarm_callback,
                self.node.connection_pool, self.node.opts.use_extensions, self.node.opts.source_version,
                self.node.opts.external_ip, self.node.opts.continent, self.node.opts.country,
                self.node.opts.should_update_source_version, self.node.blockchain_peers, self.node.account_id,
                self.node.quota_level
            )

        return self.ok({
            "removed_peer": f"{blockchain_peer_info.ip}:{blockchain_peer_info.port}"
        })



import time
from abc import ABCMeta
from dataclasses import dataclass
from typing import TYPE_CHECKING, cast, Optional, Union

from bxcommon.connections.connection_type import ConnectionType
from bxcommon.utils import convert
from bxgateway import gateway_constants
from bxcommon.utils.blockchain_utils.eth import eth_common_constants
from bxgateway.connections.abstract_blockchain_connection_protocol import AbstractBlockchainConnectionProtocol
from bxgateway.messages.eth.protocol.block_headers_eth_protocol_message import BlockHeadersEthProtocolMessage
from bxgateway.messages.eth.protocol.disconnect_eth_protocol_message import DisconnectEthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxgateway.messages.eth.protocol.hello_eth_protocol_message import HelloEthProtocolMessage
from bxgateway.messages.eth.protocol.pong_eth_protocol_message import PongEthProtocolMessage
from bxgateway.messages.eth.protocol.raw_eth_protocol_message import RawEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message import StatusEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message_v63 import StatusEthProtocolMessageV63
from bxgateway.utils.eth import frame_utils
from bxgateway.utils.eth.rlpx_cipher import RLPxCipher
from bxgateway.utils.stats.eth.eth_gateway_stats_service import eth_gateway_stats_service
from bxutils import logging

if TYPE_CHECKING:
    from bxgateway.connections.eth.eth_gateway_node import EthGatewayNode

logger = logging.get_logger(__name__)


@dataclass
class EthConnectionProtocolStatus:
    auth_message_sent: bool = False
    auth_message_received: bool = False
    auth_ack_message_sent: bool = False
    auth_ack_message_received: bool = False
    hello_message_sent: bool = False
    hello_message_received: bool = False
    status_message_sent: bool = False
    status_message_received: bool = False
    disconnect_message_received: bool = False
    disconnect_reason: Optional[int] = None


class EthBaseConnectionProtocol(AbstractBlockchainConnectionProtocol, metaclass=ABCMeta):
    node: "EthGatewayNode"

    def __init__(self, connection, is_handshake_initiator, rlpx_cipher: RLPxCipher):
        super(EthBaseConnectionProtocol, self).__init__(
            connection,
            block_cleanup_poll_interval_s=eth_common_constants.BLOCK_CLEANUP_NODE_BLOCK_LIST_POLL_INTERVAL_S
        )

        self.node = cast("EthGatewayNode", connection.node)

        self.rlpx_cipher = rlpx_cipher
        self.connection_status = EthConnectionProtocolStatus()

        self._last_ping_pong_time: Optional[float] = None
        self._handshake_complete = False

        connection.hello_messages = [
            EthProtocolMessageType.AUTH,
            EthProtocolMessageType.AUTH_ACK,
            EthProtocolMessageType.HELLO,
            EthProtocolMessageType.STATUS,
            # Ethereum Parity sends PING message before handshake is completed
            EthProtocolMessageType.PING,
            EthProtocolMessageType.DISCONNECT
        ]

        connection.message_handlers = {
            EthProtocolMessageType.AUTH: self.msg_auth,
            EthProtocolMessageType.AUTH_ACK: self.msg_auth_ack,
            EthProtocolMessageType.HELLO: self.msg_hello,
            EthProtocolMessageType.STATUS: self.msg_status,
            EthProtocolMessageType.DISCONNECT: self.msg_disconnect,
            EthProtocolMessageType.PING: self.msg_ping,
            EthProtocolMessageType.PONG: self.msg_pong,
            EthProtocolMessageType.GET_BLOCK_HEADERS: self.msg_get_block_headers
        }
        connection.pong_message = PongEthProtocolMessage(None)

        self._waiting_checkpoint_headers_request = True

        if is_handshake_initiator:
            self.connection.log_trace("Public key is known. Starting handshake.")
            self._enqueue_auth_message()
        else:
            self.connection.log_trace("Public key is unknown. Waiting for handshake request.")
            self.node.alarm_queue.register_alarm(eth_common_constants.HANDSHAKE_TIMEOUT_SEC, self._handshake_timeout)

    def msg_auth(self, msg):
        self.connection.log_trace("Beginning processing of auth message.")
        self.connection_status.auth_message_received = True
        msg_bytes = msg.rawbytes()
        decrypted_auth_msg, size = self.rlpx_cipher.decrypt_auth_message(bytes(msg_bytes))
        if decrypted_auth_msg is None:
            self.connection.log_trace("Auth message is incomplete. Waiting for more bytes.")
            return

        self.rlpx_cipher.parse_auth_message(decrypted_auth_msg)

        self._enqueue_auth_ack_message()
        self._finalize_handshake()
        self._enqueue_hello_message()
        self.connection.message_factory.reset_expected_msg_type()

    def msg_auth_ack(self, msg):
        self.connection.log_trace("Beginning processing of auth ack message.")
        self.connection_status.auth_ack_message_received = True
        auth_ack_msg_bytes = msg.rawbytes()
        self.rlpx_cipher.decrypt_auth_ack_message(bytes(auth_ack_msg_bytes))
        self._finalize_handshake()
        self._enqueue_hello_message()
        self.connection.message_factory.reset_expected_msg_type()

    def msg_ping(self, msg):
        self.connection.msg_ping(msg)
        self._last_ping_pong_time = time.time()

    def msg_pong(self, msg):
        self.connection.msg_pong(msg)
        self._last_ping_pong_time = time.time()

    def msg_hello(self, msg: HelloEthProtocolMessage):
        client_version_string = msg.get_client_version_string()
        version = msg.get_version()
        self.connection.log_info("Exchanging handshake messages with blockchain node {}, with version field {}.",
                                 client_version_string, version)
        self.connection_status.hello_message_received = True

    def msg_status(self, msg: Union[StatusEthProtocolMessage, StatusEthProtocolMessageV63]):
        self.connection.log_trace("Status message received.")
        try:
            protocol_version = msg.get_eth_version()
        except Exception:
            status_msg = StatusEthProtocolMessageV63(msg.rawbytes())
            protocol_version = status_msg.get_eth_version()
        else:
            status_msg = msg

        self.connection_status.status_message_received = True

        for peer in self.node.blockchain_peers:
            if self.node.is_blockchain_peer(self.connection.peer_ip, self.connection.peer_port):
                peer.connection_established = True

        chain_difficulty_from_status_msg = status_msg.get_chain_difficulty()
        chain_difficulty = int(self.node.opts.chain_difficulty, 16)
        fork_id = status_msg.get_fork_id()
        if isinstance(chain_difficulty_from_status_msg, int) and chain_difficulty_from_status_msg > chain_difficulty:
            chain_difficulty = chain_difficulty_from_status_msg
        self._enqueue_status_message(chain_difficulty, fork_id, protocol_version)

    def msg_disconnect(self, msg):
        self.connection_status.disconnect_message_received = True
        self.connection_status.disconnect_reason = msg.get_reason()

        self.connection.log_debug("Disconnect message was received from the blockchain node. Disconnect reason '{0}'.",
                                  self.connection_status.disconnect_reason)
        self.connection.mark_for_close()

    def msg_get_block_headers(self, msg):
        self.connection.log_trace("Replying with empty headers message to the get headers request")
        block_headers_msg = BlockHeadersEthProtocolMessage(None, [])
        self.connection.enqueue_msg(block_headers_msg)
        self._waiting_checkpoint_headers_request = False

    def get_message_bytes(self, msg):
        if isinstance(msg, RawEthProtocolMessage):
            yield msg.rawbytes()
        else:
            serialization_start_time = time.time()
            frames = frame_utils.get_frames(msg.msg_type,
                                            msg.rawbytes(),
                                            eth_common_constants.DEFAULT_FRAME_PROTOCOL_ID,
                                            eth_common_constants.DEFAULT_FRAME_SIZE)
            eth_gateway_stats_service.log_serialized_message(time.time() - serialization_start_time)

            assert frames
            self.connection.log_trace("Broke message into {} frames", len(frames))

            encryption_start_time = time.time()
            for frame in frames:
                yield self.rlpx_cipher.encrypt_frame(frame)
            eth_gateway_stats_service.log_encrypted_message(time.time() - encryption_start_time)

    def _enqueue_auth_message(self):
        auth_msg_bytes = self._get_auth_msg_bytes()
        self.connection.log_debug("Enqueued auth bytes.")
        self.connection.enqueue_msg_bytes(auth_msg_bytes)
        self.connection_status.auth_message_sent = True

    def _enqueue_auth_ack_message(self):
        auth_ack_msg_bytes = self._get_auth_ack_msg_bytes()
        self.connection.log_debug("Enqueued auth ack bytes.")
        self.connection.enqueue_msg_bytes(auth_ack_msg_bytes)
        self.connection_status.auth_ack_message_sent = True

    def _get_eth_protocol_version(self) -> int:
        for peer in self.node.blockchain_peers:
            if self.node.is_blockchain_peer(self.connection.peer_ip, self.connection.peer_port):
                return peer.blockchain_protocol_version
        return eth_common_constants.ETH_PROTOCOL_VERSION

    def _enqueue_hello_message(self):
        public_key = self.node.get_public_key()
        eth_protocol_version = eth_common_constants.ETH_PROTOCOL_VERSION
        if self.connection.CONNECTION_TYPE == ConnectionType.BLOCKCHAIN_NODE:
            eth_protocol_version = self._get_eth_protocol_version()
        elif self.connection.CONNECTION_TYPE == ConnectionType.REMOTE_BLOCKCHAIN_NODE:
            eth_protocol_version = self.node.remote_blockchain_protocol_version

        hello_msg = HelloEthProtocolMessage(
            None,
            eth_common_constants.P2P_PROTOCOL_VERSION,
            f"{gateway_constants.GATEWAY_PEER_NAME} {self.node.opts.source_version}".encode("utf-8"),
            ((b"eth", eth_protocol_version),),
            self.connection.external_port,
            public_key
        )
        self.connection.enqueue_msg(hello_msg)
        self.connection_status.hello_message_sent = True

    def _enqueue_status_message(self, chain_difficulty: int, fork_id, protocol_version: int):
        network_id = self.node.opts.network_id
        chain_head_hash = convert.hex_to_bytes(self.node.opts.genesis_hash)
        genesis_hash = chain_head_hash

        if protocol_version == gateway_constants.ETH_PROTOCOL_VERSION_63:
            status_msg = StatusEthProtocolMessageV63(
                None,
                protocol_version,
                network_id,
                chain_difficulty,
                chain_head_hash,
                genesis_hash,
            )
        else:
            status_msg = StatusEthProtocolMessage(
                None,
                protocol_version,
                network_id,
                chain_difficulty,
                chain_head_hash,
                genesis_hash,
                fork_id
            )

        self.connection.enqueue_msg(status_msg)
        self.connection_status.status_message_sent = True

    def _enqueue_disconnect_message(self, disconnect_reason):
        disconnect_msg = DisconnectEthProtocolMessage(None, [disconnect_reason])
        self.connection.enqueue_msg(disconnect_msg)

    def _get_auth_msg_bytes(self):
        auth_msg_bytes = self.rlpx_cipher.create_auth_message()
        auth_msg_bytes_encrypted = self.rlpx_cipher.encrypt_auth_message(auth_msg_bytes)

        return bytearray(auth_msg_bytes_encrypted)

    def _get_auth_ack_msg_bytes(self):
        auth_ack_msg_bytes = self.rlpx_cipher.create_auth_ack_message()
        auth_ack_msg_bytes_encrypted = self.rlpx_cipher.encrypt_auth_ack_message(auth_ack_msg_bytes)

        return bytearray(auth_ack_msg_bytes_encrypted)

    def _finalize_handshake(self):
        self._handshake_complete = True

        self.connection.log_trace("Setting up cipher.")
        self.rlpx_cipher.setup_cipher()

        self._last_ping_pong_time = time.time()
        self.node.alarm_queue.register_alarm(eth_common_constants.PING_PONG_INTERVAL_SEC, self._ping_timeout)

    def _handshake_timeout(self):
        if not self._handshake_complete:
            self.connection.log_debug("Handshake was not completed within defined timeout. Closing connection.")
            self.connection.mark_for_close()
        else:
            self.connection.log_trace("Handshake completed within defined timeout.")
        return 0

    def _ping_timeout(self) -> float:
        if not self.connection.is_alive():
            return 0

        last_ping_pong_time = self._last_ping_pong_time
        assert last_ping_pong_time is not None
        time_since_last_ping_pong = time.time() - last_ping_pong_time
        self.connection.log_trace("Ping timeout: {} seconds since last ping / pong received from node.",
                                  time_since_last_ping_pong)

        if time_since_last_ping_pong > eth_common_constants.PING_PONG_TIMEOUT_SEC:
            self.connection.log_debug("Node has not replied with ping / pong for {} seconds, more than {} limit."
                                      "Disconnecting", time_since_last_ping_pong, eth_common_constants.PING_PONG_TIMEOUT_SEC)
            self._enqueue_disconnect_message(eth_common_constants.DISCONNECT_REASON_TIMEOUT)
            self.node.alarm_queue.register_alarm(eth_common_constants.DISCONNECT_DELAY_SEC, self.connection.mark_for_close)

        return 0



from bxcommon.messages.bloxroute.block_hash_message import BlockHashMessage
from bxgateway.messages.gateway.gateway_message_type import GatewayMessageType


class BlockReceivedMessage(BlockHashMessage):
    MESSAGE_TYPE = GatewayMessageType.BLOCK_RECEIVED

    def __repr__(self):
        return "BlockReceivedMessage<block_hash: {}>".format(self.block_hash())



from dataclasses import dataclass

from bxcommon.models.config.abstract_config_model import AbstractConfigModel


@dataclass
class CronConfigModel(AbstractConfigModel):
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    throughput_stats_interval: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    info_stats_interval: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    memory_stats_interval: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    ping_interval: int = None
    # pyre-fixme[8]: Attribute has type `int`; used as `None`.
    config_update_interval: int = None



from abc import ABC

import blxr_rlp as rlp
from blxr_rlp.sedes import List

from bxcommon.messages.abstract_message import AbstractMessage


class AbstractEthMessage(AbstractMessage, ABC):
    msg_type = None
    fields = []

    # Cache of serializer instance per class
    _serializer = None

    def __init__(self, msg_bytes, *args, **kwargs):
        self._msg_bytes = None
        self._memory_view = None

        if msg_bytes is None:
            self._set_fields_from_args(args, kwargs)
            self._is_deserialized = True
        else:
            self._set_raw_bytes(msg_bytes)
            self._is_deserialized = False

    @classmethod
    def initialize_class(cls, cls_type, buf, unpacked_args):
        """
        Initialize message class with arguments. Returns cls_type instance.
        """

        return cls_type(buf)

    def serialize(self):
        encoded_payload = self._serialize_rlp_payload()
        self._set_raw_bytes(encoded_payload)

    def deserialize(self):
        assert self._msg_bytes is not None
        self._deserialize_rlp_payload(self._msg_bytes)
        self._is_deserialized = True

    def rawbytes(self) -> memoryview:
        if self._msg_bytes is None:
            self.serialize()

        assert self._msg_bytes is not None

        return self._msg_bytes

    def get_field_value(self, field_name):
        if not self._is_deserialized:
            self.deserialize()

        return getattr(self, field_name, None)

    def _serialize_rlp_payload(self):
        field_values = [getattr(self, field) for field, _ in self.fields]

        # get single value if message has just one field
        if len(field_values) == 1:
            field_values = field_values[0]

        payload = self._get_serializer().serialize(field_values)
        encoded_payload = rlp.encode(payload)

        return encoded_payload

    def _deserialize_rlp_payload(self, encoded_payload):
        if isinstance(encoded_payload, memoryview):
            encoded_payload = encoded_payload.tobytes()
        payload = rlp.decode(encoded_payload, strict=False)

        serializers = self._get_serializer()

        if serializers:
            values = serializers.deserialize(payload)

            # if message has just one field that serializers.deserialize(payload) returns a single value
            if len(self.fields) == 1:
                setattr(self, self.fields[0][0], values)
            else:
                for (field, _), value in zip(self.fields, values):
                    setattr(self, field, value)

    @classmethod
    def _get_serializer(cls):
        if cls._serializer is not None:
            return cls._serializer

        serializers = [serializer for _, serializer in cls.fields]
        serializer = serializers[0] if len(serializers) == 1 else List(serializers)

        cls._serializer = serializer

        return serializer

    def _set_raw_bytes(self, msg_bytes):
        if msg_bytes is None:
            raise ValueError("Bytes expected")

        self._msg_bytes = msg_bytes
        self._memory_view = memoryview(msg_bytes)

    def _set_fields_from_args(self, args, kwargs):
        # check keyword arguments are known
        field_set = set(field for field, _ in self.fields)

        # set positional arguments
        for (field, _), arg in zip(self.fields, args):
            setattr(self, field, arg)
            field_set.remove(field)

        # set keyword arguments, if not already set
        for (field, value) in kwargs.items():
            if field in field_set:
                setattr(self, field, value)
                field_set.remove(field)

        if len(field_set) != 0:
            raise TypeError("Not all fields initialized")

    def __eq__(self, other):
        """
        Expensive equality comparison. Use only for tests.
        """
        if not isinstance(other, AbstractEthMessage):
            return False
        else:
            return self.rawbytes() == other.rawbytes()



from typing import Dict, Any
from bxcommon import constants
from bxcommon.messages.abstract_internal_message import AbstractInternalMessage
from bxcommon.messages.bloxroute.bloxroute_message_control_flags import BloxrouteMessageControlFlags
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType


class AbstractBloxrouteMessage(AbstractInternalMessage):
    """
    Base class for internal communication between relays and gateways
    """

    MESSAGE_TYPE = BloxrouteMessageType.ABSTRACT_INTERNAL
    HEADER_LENGTH = constants.STARTING_SEQUENCE_BYTES_LEN + constants.BX_HDR_COMMON_OFF
    STARTING_BYTES_LEN = constants.STARTING_SEQUENCE_BYTES_LEN

    converted_message: Dict[int, Any]

    def __init__(self, msg_type: bytes, payload_len: int, buf: bytearray) -> None:

        super().__init__(msg_type=msg_type, payload_len=payload_len, buf=buf)

        buf[:constants.STARTING_SEQUENCE_BYTES_LEN] = constants.STARTING_SEQUENCE_BYTES

        # Control flag is set to TRUE by default
        self.set_control_flag(BloxrouteMessageControlFlags.VALID)
        self.converted_message = {}

    def get_control_flags(self) -> int:
        """
        Returns byte containing control flags for the message
        :return: byte with control flags
        """
        return self._memoryview[-1]

    def set_control_flag(self, flag: int) -> None:
        """
        Sets value to control flag
        :param flag: control flag to set
        """
        self.buf[-1] |= flag

    def remove_control_flag(self, flag: int) -> None:
        """
        Removes control flag value
        :param flag: flag value to remove
        """
        if self.buf[-1] & flag:
            self.buf[-1] ^= flag



from typing import NamedTuple, Optional
from bxcommon.connections.connection_type import ConnectionType
from bxutils import constants


class AuthenticatedPeerInfo(NamedTuple):
    connection_type: ConnectionType
    peer_id: str
    account_id: Optional[str]
    node_privileges: str = constants.DEFAULT_NODE_PRIVILEGES



import struct
from typing import Optional, Type, NamedTuple

from bxcommon import constants
from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.messages.abstract_message_factory import AbstractMessageFactory
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.abstract_broadcast_message import AbstractBroadcastMessage
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.v12.bloxroute_message_factory_v12 import bloxroute_message_factory_v12
from bxcommon.models.broadcast_message_type import BroadcastMessageType
from bxcommon.utils import crypto, uuid_pack
from bxcommon.utils.buffers.input_buffer import InputBuffer
from bxcommon.utils.object_hash import ConcatHash, Sha256Hash


class BroadcastMessagePreview(NamedTuple):
    is_full_header: bool
    block_hash: Optional[Sha256Hash]
    broadcast_type: Optional[BroadcastMessageType]
    message_id: Optional[ConcatHash]
    network_num: Optional[int]
    source_id: Optional[str]
    payload_length: Optional[int]


class _BloxrouteMessageFactoryV11(AbstractMessageFactory):
    _MESSAGE_TYPE_MAPPING = {
        # pylint: disable=protected-access
        **bloxroute_message_factory_v12._MESSAGE_TYPE_MAPPING,
        BloxrouteMessageType.REFRESH_BLOCKCHAIN_NETWORK: None,
        BloxrouteMessageType.GET_COMPRESSED_BLOCK_TXS: None,
        BloxrouteMessageType.COMPRESSED_BLOCK_TXS: None
    }

    def __init__(self) -> None:
        super(_BloxrouteMessageFactoryV11, self).__init__(self._MESSAGE_TYPE_MAPPING)

    def get_base_message_type(self) -> Type[AbstractMessage]:
        return AbstractBloxrouteMessage

    def get_broadcast_message_preview(self, input_buffer: InputBuffer) -> BroadcastMessagePreview:
        """
        Peeks the hash and network number from hashed messages.
        Currently, only Broadcast messages are supported here.
        :param input_buffer
        :return: is full header, message hash, network number, source id, payload length
        """
        # -1 for control flag length
        broadcast_header_length = self.base_message_type.HEADER_LENGTH + AbstractBroadcastMessage.PAYLOAD_LENGTH - \
                                  constants.CONTROL_FLAGS_LEN
        is_full_header = input_buffer.length >= broadcast_header_length
        if not is_full_header:
            return BroadcastMessagePreview(False, None, None, None, None, None, None)
        else:
            _is_full_message, _command, payload_length = self.get_message_header_preview_from_input_buffer(input_buffer)

            broadcast_header = input_buffer.peek_message(broadcast_header_length)

            offset = self.base_message_type.HEADER_LENGTH

            block_hash = broadcast_header[offset:offset + crypto.SHA256_HASH_LEN]
            block_hash_with_network_num = broadcast_header[offset:
                                                           offset + crypto.SHA256_HASH_LEN + constants.NETWORK_NUM_LEN]
            offset += crypto.SHA256_HASH_LEN

            network_num, = struct.unpack_from("<L", broadcast_header[offset:offset + constants.NETWORK_NUM_LEN])
            offset += constants.NETWORK_NUM_LEN

            source_id = uuid_pack.from_bytes(
                struct.unpack_from("<16s", broadcast_header[offset:offset + constants.NODE_ID_SIZE_IN_BYTES])[0])
            offset += constants.NODE_ID_SIZE_IN_BYTES

            broadcast_type_bytearray = broadcast_header[offset:offset + constants.BROADCAST_TYPE_LEN]
            broadcast_type_in_str = struct.unpack_from(
                "<4s", broadcast_type_bytearray
            )[0].decode(constants.DEFAULT_TEXT_ENCODING)
            broadcast_type = BroadcastMessageType(broadcast_type_in_str)
            message_id = ConcatHash(bytearray(block_hash_with_network_num) + broadcast_type_bytearray, 0)

            return BroadcastMessagePreview(is_full_header, Sha256Hash(block_hash), broadcast_type, message_id,
                                           network_num, source_id, payload_length)

    def __repr__(self):
        return f"{self.__class__.__name__}; message_type_mapping: {self.message_type_mapping}"


bloxroute_message_factory_v11 = _BloxrouteMessageFactoryV11()



from typing import List, Set

from bxcommon import constants
from bxcommon.feed.eth import eth_filter_handlers
from bxcommon.feed.eth.eth_raw_transaction import EthRawTransaction
from bxcommon.feed.eth.eth_transaction_feed_entry import EthTransactionFeedEntry
from bxcommon.feed.feed import Feed
from bxcommon.feed.subscriber import Subscriber

from bxutils import logging
from bxutils.logging import LogRecordType

logger = logging.get_logger(__name__)
logger_filters = logging.get_logger(LogRecordType.TransactionFiltering, __name__)


class EthTransactionProxyFeed(Feed[EthTransactionFeedEntry, EthRawTransaction]):
    def __init__(
        self,
        name: str,
        fields: List[str],
        filters: Set[str],
        network_num: int = constants.ALL_NETWORK_NUM,
    ) -> None:
        # pylint: disable=invalid-name
        self.NAME = name
        self.FIELDS = fields
        self.FILTERS = filters
        self.ALL_FIELDS = self.FIELDS

        super().__init__(self.NAME, network_num)

    def serialize(self, raw_message: EthRawTransaction) -> EthTransactionFeedEntry:
        return EthTransactionFeedEntry(
            raw_message.tx_hash,
            raw_message.tx_contents,
            raw_message.local_region,
        )

    def should_publish_message_to_subscriber(
        self,
        subscriber: Subscriber[EthTransactionFeedEntry],
        # pylint: disable=unused-argument
        raw_message: EthRawTransaction,
        serialized_message: EthTransactionFeedEntry,
    ) -> bool:
        should_publish = True
        if subscriber.filters:
            logger_filters.trace(
                "checking if should publish to {} with filters {}",
                subscriber.subscription_id,
                subscriber.filters,
            )
            contents = serialized_message.tx_contents
            state = {
                "value": eth_filter_handlers.reformat_tx_value(contents["value"]),
                "to": eth_filter_handlers.reformat_address(contents["to"]),
                "from": eth_filter_handlers.reformat_address(contents["from"]),
                "gas_price": eth_filter_handlers.reformat_gas_price(contents["gas_price"]),
                "max_priority_fee_per_gas": eth_filter_handlers.reformat_gas_price(
                    contents.get("max_priority_fee_per_gas")
                ),
                "max_fee_per_gas": eth_filter_handlers.reformat_gas_price(
                    contents.get("max_fee_per_gas")
                ),
                "method_id": eth_filter_handlers.reformat_input_to_method_id(contents["input"]),
            }
            should_publish = subscriber.validate(state)
            logger_filters.trace("should publish: {}", should_publish)
        return should_publish
